"""
Add more information to GFF annotations: gene mappings, coverage, taxonomy,
etc..

Uniprot Command
***************

If the *gene_id* of an annotation is a Uniprot ID, the script queries Uniprot
for the requested information. At the moment the information that can be added
is the taxon_id, taxon_name, lineage and mapping to EC, KO, eggNOG IDs.

It's also possible to add mappings to other databases using the *-m* option
with the correct identifier for the mapping, which can be found at `this page
<http://www.uniprot.org/faq/28>`_; for example if it's we want to add the
mappings of uniprot IDs to *BioCyc*, in the *abbreviation* column of the
mappings we find that it's identifier is *REACTOME_ID*, so we pass
*-m REACTOME* to the script (leaving *_ID* out). Mapped IDs are separated by
commas.

The taxonomy IDs are not overwritten if they are found in the annotations, the
*-f* is provided to force the overwriting of those values.

See also :ref:`gff-specs` for more informations about the GFF specifications
used.

.. note::

    As the script needs to query Uniprot a lot, it is recommended to split
    the GFF in several files, so an error in the connection doesn't waste time.

    However, a cache is kept to reduce the number of connections

Coverage Command
****************

Adds coverage information from BAM alignment files to a GFF file, using the
function :func:`mgkit.align.add_coverage_info`, the user needs to supply for
each sample a BAM file, using the `-a` option, whose parameter is in the form
`sample,samplealg.bam`. More samples can be supplied adding more `-a`
arguments.

.. hint::

    As an example, to add coverage for `sample1`, `sample2` the command line
    is::

        add-gff-info coverage -a sample1,sample1.bam -a sample2,sample2.bam \\
        inputgff outputgff

A total coverage for the annotation is also calculated and stored in the
`cov` attribute, while each sample coverage is stored into `sample_cov` as per
:ref:`gff-specs`.

Adding Coverage from samtools depth
***********************************

The *cov_samtools* allows the use of the output of *samtools* **depth**
command. The command work similarly to *coverage*, accepting compressed *depth*
files as well. If only one *depth* file is passed and no sample is passed, the
attribute in the GFF will be *cov*, otherwise the attribute will be
*sample1_cov*, *sample2_cov*, etc.

.. note::

    From version 0.4.2 the behaviour is different in the following ways:

    * the GFF file is loaded in memory to get information that helps reduce
        the memory size
    * the depth file is scanned and each time a sequence in the GFF file is
        found, coverage data is added, this way memory is freed more often
    * the GFF file is written at the very end, so chaining multiple commands
        (one for each sample) doesn't give any advantage now
    * there's no need to use the `-aa` option of `samtools depth`, since some
        assumptions are made when scanning the fils (see next point)
    * when the depth file is scanned completely, the sequences in the GFF file
        with coverage information are set coverage 0.0, instead of raising an
        error
    * average coverage is only calculated if requested, with the `-m` option

The GFF file is first loaded in memory to get information about the maximum
array size for each sequence, then the script proceeds for each depth file, to
scan it until a) all sequences in the GFF are found or b) until the end of the
file. If b) is the case, all sequences with not found in the depth file are set
to a coverage of 0.0. A warning with the number of sequences lacking coverage
is printed, for diagnostic purposes. If the the mean coverage is requested with
the `-m` option, coverage is then calculated

To create samtools *depth* files, you can use this command::

    $ samtools depth [bam_file] > [depth_file]

    where [bam_file] is the alignment to be used and [depth_file] where the
    file will be stored (by redirection). You could create a gzipped file, as
    the file can be huge (and in my experience compressing reduces by ~10x the
    file size)::

    $ samtools depth [bam_file] | gzip > [depth_file.gz]

.. warning::

    if version < 0.4.2:

    The *-aa* options must be used to pass information about all base
    pairs and sequences coverage in the BAM/SAM file.

    $ samtools depth -aa bam_file

Uniprot Offline Mappings
************************

Similar to the *uniprot* command, it uses the `idmapping <ftp://ftp.uniprot.org/pub/databases/uniprot/current_release/knowledgebase/idmapping/idmapping.dat.gz>`_
file provided by Uniprot, which speeds up the process of adding mappings and
taxonomy IDs from Uniprot gene IDs. It's not possible tough to add *EC*
mappings with this command, as those are not included in the file.

Kegg Information
****************

The *kegg* command allows to add information to each annotation. Right now the
information that can be added is restricted to the pathway(s) (reference KO) a
KO is part of and both the KO and pathway(s) descriptions. This information is
stored in keys starting with **ko_**.

Expected Aminoacidic Changes
****************************

Some scripts, like :ref:`snp-parser`, require information about the expected
number of synonymous and non-synonymous changes of an annotation. This can be
done using :meth:`mgkit.io.gff.Annotation.add_exp_syn_count` by the user of the
command `exp_syn` of this script. The attributes added to each annotation are
explained in the :ref:`gff-specs`

Adding Count Data
*****************

Count data on a per-sample basis can be added with the *counts* command. The
accepted inputs are from HTSeq-count and featureCounts. The ouput produced by
featureCounts, is the one from using its **-f** option must be used.

This script accept by default a tab separated file, with a uid in the first
column and the other columns are the counts for each sample, in the same order
as they are passed to the **-s** option. To use the featureCounts file format,
this script **-e** option must be used.

The sample names must be provided in the same order as the columns in the input
files. If the counts are FPKMS the *-f* option can be used.

Adding Taxonomy from a Table
****************************

There are cases where it may needed or preferred to add the taxonomy from a
*gene_id* already provided in the GFF file. For such cases the *addtaxa*
command can be used. It works in a similar way to the *taxonomy* command, only
it expect three different type of inputs:

    * *GI-Taxa* table from NCBI (e.g. gi_taxid_nucl.dmp, )
    * tab separated table
    * dictionary
    * HDF5

The first two are tab separated files, where on each line, the first column is
the *gene_id* that is found in the first column, while the second if the
*taxon_id*.

The third option is a serialised Python *dict*/hash table, whose keys are the
*gene_id* and the value is that gene corresponding *taxon_id*. The serialised
formats accepted are msgpack, json and pickle. The *msgpack* module must be
importable. The option to use json and msgpack allow to integrate this script
with other languages without resorting to a text file.

The last option is a HDF5 created using the *to_hdf* command in
:ref:`taxon-utils`. This requires `pandas` installed and `pytables` and it
provides faster lookup of IDs in the table.

While the default is to look for the *gene_id* attribute in the GFF annotation,
another attribute can be specified, using the **-gene-attr** option.

.. note::

    the dictionary content is loaded after the table files and its keys and
    corresponding values takes precedence over the text files.

.. warning::

    from September 2016 NCBI will retire the GI. In that case the same
    kind of table can be built from the *nucl_gb.accession2taxid.gz* file
    The format is different, but some information can be found in
    :func:`mgkit.io.blast.parse_accession_taxa_table`


Adding information from Pfam
****************************

Adds the Pfam description for the annotation, by downloading the list from
Pfam.

The options allow to specify in which attribute the ID/ACCESSION is stored
(defaults to *gene_id*) and which one between ID/ACCESSION is the value of that
attribute (defaults to *ID*). if no description is found for the family, a
warning message is logged.

Changes
*******

.. versionchanged:: 0.4.2
    added *-m* option for *cov_samtools* command, to calculate the average
    coverage for an annotation (*cov* attribute). Fixed loading compressed
    files

.. versionchanged:: 0.3.4
    removed the *taxonomy* command, since a similar result can be obtained with
    *taxon-utils lca* and *add-gff-info addtaxa*. Removed *eggnog* command and
    added option to verbose the logging in *cov_samtools* (now is quiet), also
    changed the interface

.. versionchanged:: 0.3.3
    changed how *addtaxa* *-a* works, to allow the use of *seq_id* as key to
    add the taxon_id

.. versionchanged:: 0.3.0
    added *cov_samtools* command, *--split* option to *exp_syn*, *-c* option to
    *addtaxa*. *kegg* now does not skip annotations when the attribute is not
    found.

.. versionchanged:: 0.2.6
    added *skip-no-taxon* option to *addtaxa*

.. versionchanged:: 0.2.5
    if a dictionary is supplied to *addtaxa*, the GFF is not preloaded

.. versionchanged:: 0.2.3
    added *pfam* command, renamed *gitaxa* to *addtaxa* and made it general

.. versionchanged:: 0.2.2
    added *eggnog*, *gitaxa* and *counts* command

.. versionchanged:: 0.2.1

* added *-d* to *uniprot* command
* added cache to *uniprot* command
* added *kegg* command (cached)

.. versionchanged:: 0.1.16
    added *exp_syn* command

.. versionchanged:: 0.1.15
    *taxonomy* command *-b* option changed

.. versionchanged:: 0.1.13

* added *--force-taxon-id* option to the *uniprot* command
* added *coverage* command
* added *taxonomy* command
* added *unipfile* command

.. versionadded:: 0.1.12
"""
from __future__ import division
from builtins import zip
from future.utils import viewitems
import logging
import itertools
import pysam
import json
import pickle
from tqdm import tqdm
import click
import msgpack
import mgkit
import mgkit.counts.func
from . import utils
from .. import align
from .. import logger
from .. import taxon
from .. import kegg
from ..io import gff, blast, fasta, open_file
from ..io import uniprot as uniprot_io
from ..net import uniprot as uniprot_net
from ..net import pfam
from ..utils.dictionary import cache_dict_file, HDFDict


LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('kegg', help="""Adds information and mapping from Kegg""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-c', '--email', help='Contact email', required=True)
@click.option('-d', '--description', default=False, is_flag=True,
              help='Add Kegg description')
@click.option('-p', '--pathways', is_flag=True, default=False,
              help='Add pathways ID involved')
@click.option('-m', '--kegg-id', default='gene_id',
              help="""In which attribute the Kegg ID is stored (defaults to *gene_id*)""")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def kegg_command(verbose, email, description, pathways, kegg_id, input_file,
                 output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    kegg_client = kegg.KeggClientRest()

    LOG.info('Retrieving KO names')
    ko_names = kegg_client.get_ids_names('ko')

    if pathways:
        LOG.info('Retrieving Pathways names')
        # Changes the names of the keys to *ko* instead of *map*
        path_names = dict(
            (path_id.replace('map', 'ko'), name)
            for path_id, name in viewitems(kegg_client.get_ids_names('path'))
        )

    ko_cache = {}

    for annotation in gff.parse_gff(input_file, gff_type=gff.from_gff):
        try:
            ko_id = annotation.attr[kegg_id]
        except KeyError:
            annotation.to_file(output_file)
            continue
        # if more than one KO is defined
        if ',' in ko_id:
            LOG.warning(
                'More than one KO assigned, skipping annotation: %s',
                annotation.uid
            )
            annotation.to_file(output_file)
            continue
        try:
            ko_info = ko_cache[ko_id]
        except KeyError:
            ko_info = {}
            if pathways:
                ko_info['pathway'] = kegg_client.link_ids('path', ko_id)
                # If left empty, no pathway will be saved
                if ko_info['pathway']:
                    ko_info['pathway'] = list(
                        path_id
                        for path_id in ko_info['pathway'][ko_id]
                        if path_id.startswith('ko')
                    )
            ko_cache[ko_id] = ko_info
        if description:
            annotation.attr['ko_description'] = ko_names.get(ko_id, ko_id)
        if pathways and ko_info['pathway']:
            annotation.attr['ko_pathway'] = ','.join(ko_info['pathway'])
            annotation.attr['ko_pathway_names'] = ','.join(
                path_names[path_id]
                for path_id in ko_info['pathway']
            )

        annotation.to_file(output_file)


def add_uniprot_info(annotations, email, force_taxon_id, taxon_id, lineage,
                     eggnog, enzymes, kegg_orthologs, protein_names, mapping,
                     info_cache):
    columns = []

    if taxon_id:
        columns.append('organism')
        columns.append('organism-id')
    if eggnog:
        columns.append('database(EGGNOG)')
    if kegg_orthologs:
        columns.append('database(KO)')
    if enzymes:
        columns.append('ec')
    if lineage:
        columns.append('lineage(ALL)')
    if mapping is not None:
        for db in mapping:
            columns.append('database({0})'.format(db))
    if protein_names:
        columns.append('protein names')

    if not columns:
        return

    LOG.info("Retrieving gene information from Uniprot")

    gene_ids = set(
        x.gene_id
        for x in annotations
        if x.gene_id not in info_cache
    )

    # It may be that all gene_ids are already cached. In which case there's no
    # point querying Uniprot (it will probably give an error)
    if gene_ids:
        data = uniprot_net.get_gene_info(
            gene_ids,
            columns=columns,
            contact=email
        )

        info_cache.update(data)

    for annotation in annotations:
        try:
            gene_info = info_cache[annotation.gene_id]
        except KeyError:
            # no data was found
            continue

        for column, values in viewitems(gene_info):
            # nothing found
            if not values:
                continue

            if column == 'organism-id':
                if (annotation.taxon_id and force_taxon_id) or \
                   (annotation.taxon_id is None):
                    annotation.attr['taxon_id'] = int(values)
                    annotation.attr['taxon_db'] = 'UNIPROT'
                    # test with a try/expect maybe
                    if 'organism' in columns:
                        annotation.attr['taxon_name'] = gene_info['organism']
            elif column.startswith('lineage'):
                annotation.attr['lineage'] = gene_info['lineage(ALL)']
            elif column.startswith('database'):
                annotation.attr[
                    'map_{0}'.format(
                        column[:-1].split('(')[1]
                    )
                ] = ','.join(values)
            elif column == 'ec':
                if isinstance(values, list):
                    annotation.attr['EC'] = ','.join(
                        x.split('-')[0]
                        for x in values
                    )
                else:
                    annotation.attr['EC'] = values
            elif column == 'protein names':
                annotation.attr['uniprot_description'] = values


@main.command('uniprot', help="""Adds information from GFF whose gene_id is
              from Uniprot""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-c', '--email', help='Contact email', required=True)
@click.option('--buffer', type=click.INT, default=50, show_default=True,
              help='Number of annotations to keep in memory')
@click.option('-f', '--force-taxon-id', default=False, is_flag=True,
              help='Overwrite taxon_id if already present')
@click.option('-t', '--taxon-id', default=False, is_flag=True,
              help="""Add taxonomic ids to annotations, if taxon_id is found, it won't be Overwritten.""")
@click.option('-l', '--lineage', default=False, is_flag=True,
              help='Add taxonomic lineage to annotations')
@click.option('-e', '--eggnog', default=False, is_flag=True,
              help='Add eggNOG mappings to annotations')
@click.option('-ec', '--enzymes', default=False, is_flag=True,
              help='Add EC mappings to annotations')
@click.option('-ko', '--kegg_orthologs', default=False, is_flag=True,
              help='Add KO mappings to annotations')
@click.option('-d', '--protein-names', default=False, is_flag=True,
              help='Add Uniprot description')
@click.option('-m', '--mapping', multiple=True, default=None,
              help='Add any DB mappings to annotations')
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def uniprot_command(verbose, email, buffer, force_taxon_id, taxon_id, lineage,
                    eggnog, enzymes, kegg_orthologs, protein_names, mapping,
                    input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    if buffer < 1:
        buffer = 1

    ann_buffer = []

    info_cache = {}

    for annotation in gff.parse_gff(input_file, gff_type=gff.from_gff):

        ann_buffer.append(annotation)

        if len(ann_buffer) == buffer:

            add_uniprot_info(ann_buffer, email, force_taxon_id, taxon_id,
                             lineage, eggnog, enzymes, kegg_orthologs,
                             protein_names, mapping, info_cache)

            for annotation in ann_buffer:
                annotation.to_file(output_file)

            ann_buffer = []
    else:
        add_uniprot_info(ann_buffer, email, force_taxon_id, taxon_id, lineage,
                         eggnog, enzymes, kegg_orthologs, protein_names,
                         mapping, info_cache)

        for annotation in ann_buffer:
            annotation.to_file(output_file)


def split_sample_alg(ctx, param, values):
    "Split sample/alignment option"

    new_values = []
    for value in values:
        try:
            sample, bam_file_name = value.split(',', 1)
        except ValueError:
            raise click.BadParameter(
                "Can't get get both sample and bam file from '%s'" % value
            )
        new_values.append((sample, bam_file_name))

    return new_values


@main.command('coverage', help="""Adds coverage information from BAM Alignment
              files""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-a', '--sample-alignment', multiple=True, required=True,
              callback=split_sample_alg,
              help='sample name and correspondent alignment file separated by comma')
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def coverage_command(verbose, sample_alignment, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)
    samples = []
    bam_files = []

    for sample, bam_file_name in sample_alignment:
        samples.append(sample)
        bam_files.append(pysam.Samfile(bam_file_name, 'rb'))

    if len(samples) != len(set(samples)):
        utils.exit_script("There are duplicate sample names", 1)

    annotations = list(gff.parse_gff(input_file))

    align.add_coverage_info(annotations, bam_files, samples)

    gff.write_gff(annotations, output_file)


@main.command('exp_syn', help="""Adds expected synonymous and non-synonymous
              changes information""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-r', '--reference', required=True, type=click.File('rb', lazy=False),
              help='reference sequence in fasta format')
@click.option('-s', '--split', default=False, is_flag=True,
              help='''Split the sequence header of the reference at the first space, to emulate BLAST behaviour''')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def exp_syn_command(verbose, reference, split, progress, input_file,
                    output_file):
    """
    .. versionadded:: 0.1.16
    """
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    seqs = dict(
        (
            seq_id.split(' ')[0] if split else seq_id,
            seq
        )
        for seq_id, seq in fasta.load_fasta(reference)
    )

    iterator = gff.parse_gff(input_file)

    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        curr_seq = seqs[annotation.seq_id]
        if len(curr_seq) < annotation.end:
            LOG.error("Annotation (uid %s) is outside the the sequence (sequence length %d): %s", annotation.uid, len(curr_seq), annotation)
        annotation.add_exp_syn_count(curr_seq)

        annotation.to_file(output_file)


@main.command('unipfile', help="""Adds expected synonymous and non-synonymous
              changes information""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-i', '--mapping-file', default='idmapping.dat.gz', required=True,
              help="Uniprot mapping file")
@click.option('-f', '--force-taxon-id', default=False, is_flag=True,
              help="Overwrite taxon_id if already present")
@click.option('-m', '--mapping', multiple=True, default=None, required=True,
              type=click.Choice(list(uniprot_io.MAPPINGS.values())),
              help="Mappings to add")
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def uniprot_offline_command(verbose, mapping_file, force_taxon_id, mapping,
                            progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    LOG.info("Mappings selected: %s", ', '.join(mapping))

    annotations = []
    gene_ids = set()

    for annotation in gff.parse_gff(input_file):
        annotations.append(annotation)
        gene_ids.add(annotation.gene_id)

    iterator = uniprot_io.uniprot_mappings_to_dict(
        mapping_file,
        gene_ids=set(gene_ids),
        mappings=set(mapping),
        num_lines=None
    )

    if progress:
        iterator = tqdm(iterator)

    file_mappings = dict(iterator)

    count = 0

    for annotation in annotations:
        try:
            mappings = file_mappings[annotation.gene_id]
            for mapping_id, mapping_values in viewitems(mappings):
                if mapping_id == uniprot_io.MAPPINGS['taxonomy']:
                    if (annotation.taxon_id and force_taxon_id) or \
                       (annotation.taxon_id is None):
                        annotation.taxon_id = mapping_values[0]
                        annotation.taxon_db = 'UNIPROT'
                else:
                    annotation.set_mapping(mapping_id, mapping_values)
            count += 1
        except KeyError:
            pass

        annotation.to_file(output_file)

    LOG.info(
        "Number of annotation changed: %d/%d (%.2f%%)",
        count,
        len(annotations),
        count / len(annotations) * 100
    )


def load_htseq_count_files(count_files, samples):
    counts = {}

    for sample, count_file in zip(samples, count_files):
        for uid, count in mgkit.counts.func.load_htseq_counts(count_file):
            if uid not in counts:
                counts[uid] = {}
            counts[uid][sample] = count

    return counts


def load_featurecounts_files(count_files, samples):
    counts = {}
    for sample, count_file in zip(samples, count_files):
        for line in open_file(count_file, 'rb'):
            line = line.decode('ascii')
            if line.startswith('#') or line.lower().startswith('geneid'):
                continue
            line = line.strip().split('\t')
            uid = line[0]
            counts[uid] = dict(zip(samples, line[6:]))
    return counts


@main.command('counts', help="""Adds counts data to the GFF file""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-s', '--samples', required=True, multiple=True,
              help="""Sample names, in the same order as the count files""")
@click.option('-c', '--count-files', multiple=True, required=True, help="Count file(s)")
@click.option('-f', '--fpkms', default=False, is_flag=True,
              help="If the counts are FPKMS")
@click.option('-e', '--featureCounts', default=False, is_flag=True,
              help="""If the counts files are from featureCounts""")
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def counts_command(verbose, samples, count_files, fpkms, featurecounts,
                   progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    LOG.info("Samples passed: %s", ','.join(samples))

    if featurecounts:
        counts = load_featurecounts_files(count_files, samples)
    else:
        counts = load_htseq_count_files(count_files, samples)

    key = "fpkms_{}" if fpkms else "counts_{}"

    iterator = gff.parse_gff(input_file)

    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        try:
            ann_counts = counts[annotation.uid]
        except KeyError:
            LOG.warning("No counts found for annotation %s", annotation.uid)
            annotation.to_file(output_file)
            continue

        for sample in samples:
            annotation.attr[key.format(sample)] = ann_counts.get(sample, 0)

        annotation.to_file(output_file)


def parse_hdf5_arg(ctx, param, values):
    if values is None:
        return values
    try:
        file_name, table = values.strip().split(':')
    except ValueError:
        raise click.BadParameter("""The HDF5 file name must be followed by ':'
                                 and the name of the table""")

    return (file_name, table)


@main.command('addtaxa', help='''Adds taxonomy information from a GI-Taxa,
              gene_id/taxon_id table or a dictionary serialised as a
              pickle/msgpack/json file, or a table in a HDF5 file''')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--gene-taxon-table', default=None,
              help="""GIDs taxonomy table (e.g. gi_taxid_nucl.dmp.gz) or a similar file where GENE/TAXON are tab separated and one per line""")
@click.option('-a', '--gene-attr', default='gene_id',
              help="""In which attribute the GENEID in the table is stored (defaults to *gene_id*)""")
@click.option('-f', '--hdf-table', default=None, callback=parse_hdf5_arg,
              help="""HDF5 file and table name to use for taxon_id lookups. The format to pass is the file name, colon and the table file hf5:taxa-table. The index in the table is the accession_id, while the column `taxon_id` stores the taxon_id as int""")
@click.option('-x', '--taxonomy', default=None, help="""Taxonomy file - If given, both *taxon_name* and *lineage* attributes will be set""")
@click.option('-d', '--dictionary', default=None, help="""A serialised dictionary, where the key is the GENEID and the value is TAXONID. It can be in json or msgpack format (can be a compressed file) *Note*: the dictionary values takes precedence over the table files""")
@click.option('-e', '--skip-no-taxon', default=False, is_flag=True,
              help="""If used, annotations with no taxon_id won't be outputted""")
@click.option('-db', '--taxon-db', default='NONE', help="""DB used to add the taxonomic information""")
@click.option('-c', '--cache-table', default=False, is_flag=True,
              help="""If used, annotations are not preloaded, but the taxa table is cached, instead""")
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def addtaxa_command(verbose, gene_taxon_table, hdf_table, gene_attr, taxonomy,
                    dictionary, skip_no_taxon, taxon_db, cache_table,
                    progress, input_file, output_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    if (gene_taxon_table is not None) and (not cache_table):
        annotations = []
        gene_ids = set()

        # the annotations are preloaded if a table is supplied, such as
        # the one from NCBI, so only the gene_ids necessary are taken from that
        # table to save memory

        for annotation in gff.parse_gff(input_file):
            annotations.append(annotation)
            gene_ids.add(annotation.get_attr(gene_attr, str))
        gene_ids = dict(
            blast.parse_accession_taxa_table(
                gene_taxon_table,
                acc_ids=gene_ids,
                key=0,
                value=1,
                no_zero=True
            )
        )
    elif (gene_taxon_table is not None) and cache_table:
        annotations = gff.parse_gff(input_file)
        gene_ids = cache_dict_file(
            blast.parse_accession_taxa_table(
                gene_taxon_table,
                key=0,
                value=1,
                no_zero=True
            )
        )
    elif hdf_table is not None:
        try:
            gene_ids = HDFDict(hdf_table[0], hdf_table[1])
        except ValueError as e:
            utils.exit_script(
                str(e),
                3
            )
        annotations = gff.parse_gff(input_file)
    else:
        # in case a dictionary is supplied, it's expected to fit in memory,
        # meaning that the GFF doesn't have to be preloaded
        annotations = gff.parse_gff(input_file)
        gene_ids = {}

    if (dictionary is not None) and (hdf_table is None):
        dict_file = open_file(dictionary, 'r')
        if '.json' in dictionary:
            gene_ids.update(json.load(dict_file))
        elif '.msgpack':
            gene_ids.update(msgpack.load(dict_file))
        else:
            gene_ids.update(pickle.load(dict_file))

        # Ensures all taxon_ids are *int*
        gene_ids = dict(
            (key, int(value))
            for key, value in viewitems(gene_ids)
        )

    if taxonomy is not None:
        taxonomy = taxon.Taxonomy(taxonomy)

    if progress:
        annotations = tqdm(annotations)

    count = 0
    for lineno, annotation in enumerate(annotations):
        gene_id = annotation.get_attr(gene_attr, str)
        try:
            taxon_id = gene_ids[gene_id]
            annotation.taxon_id = taxon_id
            annotation.set_attr('taxon_db', taxon_db)
            count += 1
        except KeyError:
            LOG.warning(
                "No Taxon ID for GENE - %s",
                gene_id
            )
            taxon_id = None
            if skip_no_taxon:
                continue

        if (taxonomy is not None) and (taxon_id is not None):
            try:
                annotation.attr['taxon_name'] = taxonomy[taxon_id].s_name
                annotation.attr['lineage'] = ','.join(
                    taxon_name
                    for taxon_name in taxon.get_lineage(taxonomy, taxon_id, names=True, only_ranked=True)
                    if taxon_name
                )
            except KeyError:
                LOG.warning("Taxon ID %d not found in the Taxonomy", taxon_id)

        annotation.to_file(output_file)

    LOG.info("Percentage of annotations changed %.2f%%", count / (count + 1) * 100)


@main.command('pfam', help="""Adds information from Pfam""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-i', '--id-attr', default='gene_id', help="""In which attribute the Pfam ID/ACCESSION is stored (defaults to *gene_id*)""")
@click.option('-a', '--use-accession', default=False, is_flag=True, help="""If used, the attribute value is the Pfam ACCESSION (e.g. PF06894), not ID (e.g. Phage_TAC_2)""")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def pfam_command(verbose, id_attr, use_accession, input_file, output_file):
    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    LOG.info("Downloading Pfam Information")

    pfam_families = pfam.get_pfam_families(
        'acc' if use_accession else 'id'
    )

    for annotation in gff.parse_gff(input_file):
        pfam_id = annotation.attr[id_attr]
        try:
            annotation.attr['pfam_description'] = pfam_families[pfam_id][1]
        except KeyError:
            LOG.warning("No description found for family %s", pfam_id)
        annotation.to_file(output_file)


def update_annotation_coverage(annotations, depth, sample):
    for annotation in annotations:
        cov = depth.region_coverage(
            annotation.seq_id,
            annotation.start,
            annotation.end
        )
        if sample is None:
            annotation.set_attr('cov', cov)
        else:
            annotation.set_attr(sample, cov)


@main.command('cov_samtools', help="""Adds information from samtools_depth""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-m', '--average', is_flag=True, help='if one or more samples are provided, the average coverage is calculated')
@click.option('-s', '--samples', default=None, multiple=True,
              help='''Sample name, will add a `sample_cov` in the GFF file. If not passed, the attribute will be `cov`''')
@click.option('-d', '--depths', required=True, multiple=True,
              help='`samtools depth -aa` file')
@click.option('-n', '--num-seqs', default=0,  type=click.INT,
              show_default=True, help='''Number of sequences to update the log. If 0, no message is logged''')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('input-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def samtools_depth_command(verbose, average, samples, depths, num_seqs, progress,
                           input_file, output_file):
    logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if not samples:
        samples = (None,)
    else:
        samples = ['{}_cov'.format(sample) for sample in samples]

    max_size_dict = {}
    annotations = gff.group_annotations(gff.parse_gff(input_file), key_func=lambda x: x.seq_id)

    if progress:
        it = tqdm(annotations.items(), desc='Max lengths')
    else:
        it = annotations.items()
    for seq_id, seq_annotations in it:
        max_size_dict[seq_id] = max(annotation.end for annotation in seq_annotations)

    depths = [
        align.SamtoolsDepth(
            file_name,
            num_seqs=None if num_seqs == 0 else num_seqs,
            max_size_dict=max_size_dict,
        )
        for file_name in depths
    ]
    if len(samples) != len(depths):
        utils.exit_script('Number of samples different from number of files', 2)

    if progress and (len(samples) > 1):
        depth_it = tqdm(zip(samples, depths), desc='Samples', total=len(samples))
    else:
        depth_it = zip(samples, depths)

    for sample, depth in depth_it:
        seq_found = set()

        if progress:
            seq_bar = tqdm(desc='Sequences', total=len(annotations))
        while True:
            seq_id = depth.advance_file()
            # sequence is in the annotations
            if seq_id in annotations:
                update_annotation_coverage(annotations[seq_id], depth, sample)
                seq_found.add(seq_id)
                depth.drop_sequence(seq_id)
                if progress:
                    seq_bar.update(n=1)
            # reached the end of the file
            # updated the remaining sequences
            elif seq_id is None:
                for not_found in set(annotations) - seq_found:
                    update_annotation_coverage(annotations[not_found], depth,
                                               sample)
                    if progress:
                        seq_bar.update(n=1)
                LOG.warning(
                    'Cannot find %d of %d sequences in sample %s depth file',
                    len(depth.not_found), len(annotations), sample)
                break
            # found all sequences, no need to continue scanning the file
            elif len(seq_found) == len(annotations):
                LOG.info("Found all sequences in GFF file (sample %s)", sample)
                break
            # the sequence was not in the annotations, skip
            # continue scanning the depth file and drop the seq_id
            # (redundat?)
            elif seq_id not in annotations:
                depth.drop_sequence(seq_id)

            # some debug code, to check the density of the SparseArray(s)
            # if (num_seqs > 0) and (len(depth.density) % num_seqs == 0):
            #     LOG.debug(sum(depth.density) / len(depth.density))

        # closes the progress bar
        if progress:
            seq_bar.close()

    # If no sample was provided it's assumed that the average was calculated
    if average and (samples[0] is not None):
        max_coverage = 0

        LOG.info('Calculating average coverage')
        for seq_annotations in annotations.values():
            for annotation in seq_annotations:
                d = annotation.sample_coverage
                annotation.set_attr('cov', sum(d.values()) / len(d))
                max_coverage = max(max_coverage, max(d.values()))
        LOG.info('Maximum Coverage found: %d', max_coverage)

    LOG.info('Writing annotations to file (%r)', output_file)
    for annotation in itertools.chain(*annotations.values()):
        annotation.to_file(output_file)



import pytest
from mgkit import kegg
from conftest import skip_no_connection

@pytest.fixture
def keggclient():
    return kegg.KeggClientRest()

pytestmark = skip_no_connection

def test_keggclient_link1(keggclient):
    assert keggclient.link_ids('rn', 'K00201') == {'K00201': ['R03015', 'R08060']}


def test_keggclient_list1(keggclient):
    assert 'cpd:C20660\tWybutosine in tRNA(Phe)\n' in keggclient.list_ids('cpd')


def test_keggclient_get1(keggclient):
    assert 'DBLINKS     PubChem: 172232382' in keggclient.get_entry('cpd:C20660')


def test_keggclient_get_names1(keggclient):
    assert 'M00002' in keggclient.get_ids_names('module')


def test_keggclient_get_names2(keggclient):
    assert 'md:M00002' in keggclient.get_ids_names('module', strip=False)


def test_cache_io(keggclient, tmpdir):
    query = keggclient.link_ids('rn', 'K00201')['K00201']
    file_name = tmpdir.join('cache.pickle').strpath
    keggclient.write_cache(file_name)

    cached = kegg.KeggClientRest(file_name)
    assert cached.cache == keggclient.cache


def test_cache_link_ids1(keggclient):
    query = keggclient.link_ids('rn', 'K00201')['K00201']
    assert query == keggclient.cache['link_ids']['rn']['K00201']


def test_find(keggclient):
    assert keggclient.find('CH4', 'compound') == {'C01438': 'Methane; CH4'}


def test_conv(keggclient):
    assert 'b0217' in keggclient.conv('ncbi-geneid', 'eco')


def test_get_ids_names(keggclient):
    assert 'K00001' in keggclient.get_ids_names('ko')



"""
GFF filtering
"""
from __future__ import division
import itertools


def choose_annotation(ann1, ann2, overlap=100, choose_func=None):
    """
    .. versionadded:: 0.1.12

    Given two :class:`mgkit.io.gff.Annotation`, if one of of the two
    annotations either is contained in the other or they overlap for at least a
    *overlap* number of bases, *choose_func* will be applied to both. The
    result of *choose_func* is the the annotation to be discarderd. It returns
    *None* if the annotations should be both kept.

    No checks are made to ensure that the two annotations are on the same
        sequence and strand, as the *intersect* method of
        :class:`mgkit.io.gff.Annotation` takes care of them.

    Arguments:
        ann1: instance of :class:`mgkit.io.gff.Annotation`
        ann2: instance of :class:`mgkit.io.gff.Annotation`
        overlap (int, float): number of bases overlap that trigger the
            filtering
        choose_func (None, func): function that accepts *ann1* and *ann2* and
            return the one to be discarded or None if both are accepted

    Returns:
        (None, Annotation): returns either the :class:`mgkit.io.gff.Annotation`
        to be discarded or None, which is the result of *choose_func*

    .. note::

        If *choose_func* is *None*, the default function is used::

            lambda a1, a2: min(a1, a2, key=lambda el: (el.dbq, el.bitscore,
                               len(el)))

        In order of importance the db quality, the bitscore and the length. The
        annotation with the lowest tuple value is the one to discard.

    """

    if choose_func is None:
        def choose_func(a1, a2):
            return min(a1, a2, key=lambda el: (el.dbq, el.bitscore, len(el)))

    intersect = ann1.intersect(ann2)

    if intersect is not None:
        # if the intersection is the same size of one of the annotations size,
        # it means that one contain the other
        if (len(intersect) == len(ann1)) or (len(intersect) == len(ann2)):
            return choose_func(ann1, ann2)
        else:
            # if the overlap is longer than the threshold
            if len(intersect) > overlap:
                return choose_func(ann1, ann2)

    return None


def filter_annotations(annotations, choose_func=None, sort_func=None,
                       reverse=True):
    """
    .. versionadded:: 0.1.12

    Filter an iterable of :class:`mgkit.io.gff.Annotation` instances sorted
    using *sort_func* as key in *sorted* and if the order is to be *reverse*;
    it then applies *choose_func* on all possible pair combinations, using
    itertools.combinations.

    By default *choose_func* is :func:`choose_annotation` with the default
    values, the list of annotation is sorted by bitscore, from the highest to
    the lowest value.

    Arguments:
        annotations (iterable): iterable of :class:`mgkit.io.gff.Annotation`
            instances
        choose_func (func, None): function used to select the *losing*
            annotation; if `None`, it will be :func:`choose_annotation` with
            default values
        sort_func (func, None): by default the sorting key is the bitscore of
            the annotations
        reverse (bool): passed to `sorted`, by default is reversed

    Returns:
        set: a set with the annotations that pass the filtering
    """

    if choose_func is None:
        choose_func = choose_annotation

    if sort_func is None:
        def sort_func(x): return x.bitscore

    annotations = sorted(annotations, key=sort_func, reverse=reverse)

    to_remove = set()

    for ann1, ann2 in itertools.combinations(annotations, 2):

        if ann1 in to_remove or ann2 in to_remove:
            continue

        to_remove.add(choose_func(ann1, ann2))

    return set(annotations) - to_remove


def filter_base(annotation, attr=None, value=None):
    """
    Checks if an annotation attribute is equal to the requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        attr (str): attribute of the annotation
        value: the value that the attribute should be equal to

    Returns:
        bool: True if the supplied value is equal to the attribute ot False
        otherwise
    """
    return getattr(annotation, attr, None) == value


def filter_len(annotation, value=None, greater=True):
    """
    Checks if an annotation length is longer, equal of shorter than the
    requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        value (int): the length to which the attribute should be compared to
        greater (bool): if True the annotation length must be equal or greater
            than and if False equal of lower than

    Returns:
        bool: True if the test passes
    """
    if greater:
        return len(annotation) >= value
    else:
        return len(annotation) <= value


def filter_base_num(annotation, attr=None, value=None, greater=True):
    """
    Checks if an annotation attribute is greater, equal of lower than the
    requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        attr (str): attribute of the annotation
        value (int): the value to which the attribute should be compared to
        greater (bool): if True the attribute value must be equal or greater
            than and if False equal of lower than

    Returns:
        bool: True if the test passes
    """
    annotation_value = getattr(annotation, attr, None)
    if annotation_value is None:
        return False
    annotation_value = float(annotation_value)
    if greater:
        return annotation_value >= value
    else:
        return annotation_value <= value


def filter_attr_num(annotation, attr=None, value=None, greater=True):
    """
    Checks if an annotation *attr* dictionary contains a key whose value is
    greater than or equal, or lower than or equal, for the requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        attr (str): key in the :attr:`mgkit.io.gff.Annotation.attr`
            dictionary
        value (int): the value to which we need to compare
        greater (bool): if True the value must be equal or greater than and if
            False equal of lower than

    Returns:
        bool: True if the test passes
    """
    annotation_value = annotation.attr.get(attr, None)
    if annotation_value is None:
        return False
    annotation_value = float(annotation_value)
    if greater:
        return annotation_value >= value
    else:
        return annotation_value <= value


def filter_attr_num_s(annotation, attr=None, value=None, greater=True):
    """
    .. versionadded:: 0.3.1

    Checks if an annotation *attr* dictionary contains a key whose value is
    greater or lower than the requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        attr (str): key in the :attr:`mgkit.io.gff.Annotation.attr`
            dictionary
        value (int): the value to which we need to compare
        greater (bool): if True the value must be greater than and if
            False lower than

    Returns:
        bool: True if the test passes
    """
    annotation_value = annotation.attr.get(attr, None)
    if annotation_value is None:
        return False
    annotation_value = float(annotation_value)
    if greater:
        return annotation_value > value
    else:
        return annotation_value < value


def filter_attr_str(annotation, attr=None, value=None, equal=True):
    """
    Checks if an annotation *attr* dictionary contains a key shose value is
    equal to, or contains the requested value

    Arguments:
        annotation: :class:`mgkit.io.gff.Annotation` instance
        attr (str): key in the :attr:`mgkit.io.gff.Annotation.attr`
            dictionary
        value (int): the value to which we need to compare
        equal (bool): if True the value must be equal and if False equal value
            must be contained

    Returns:
        bool: True if the test passes
    """
    annotation_value = annotation.attr.get(attr, None)
    if annotation_value is None:
        return False
    if equal:
        return annotation_value == value
    else:
        return value in annotation_value



import pytest
from mgkit.io import fastq
from mgkit.io import open_file

@pytest.fixture
def fastq_file(shared_datadir):
    return str(shared_datadir / 'SP1.fq')


def test_load_fastq1(fastq_file):

    assert sum(1 for record in fastq.load_fastq(fastq_file)) == 250


def test_load_fastq2(fastq_file):

    header, seq, qual = next(fastq.load_fastq(fastq_file))

    assert seq == 'TTTCCGGGGCACATAATCTTCAGCCGGGCGC'


def test_load_fastq3(fastq_file):

    header, seq, qual = next(fastq.load_fastq(fastq_file))

    assert header == 'cluster_2:UMI_ATTCCG'


def test_load_fastq4(fastq_file):

    header, seq, qual = next(fastq.load_fastq(fastq_file))

    assert qual == '9C;=;=<9@4868>9:67AA<9>65<=>591'


def test_load_fastq4(fastq_file):

    header, seq, qual = next(fastq.load_fastq(fastq_file, num_qual=True))

    assert list(qual) == [24, 34, 26, 28, 26, 28, 27, 24, 31, 19, 23, 21, 23, 29, 24, 25, 21, 22, 32, 32, 27, 24, 29, 21, 20, 27, 28, 29, 20, 24, 16]


def test_write_fastq1(fastq_file, tmpdir):

    header, seq, qual = next(fastq.load_fastq(fastq_file))

    file_name = (tmpdir / 'test.fq').strpath

    file_handle = open_file(file_name, 'w')

    fastq.write_fastq_sequence(file_handle, header, seq, qual)
    file_handle.close()

    headerw, seqw, qualw = next(fastq.load_fastq(file_name))

    assert (header, seq, qual) == (headerw, seqw, qualw)


def test_write_fastq2(fastq_file, tmpdir):

    header, seq, qual = next(fastq.load_fastq(fastq_file, num_qual=True))

    file_name = (tmpdir / 'test.fq').strpath

    file_handle = open_file(file_name, 'w')

    fastq.write_fastq_sequence(file_handle, header, seq, qual)
    file_handle.close()

    headerw, seqw, qualw = next(fastq.load_fastq(file_name, num_qual=True))

    assert (header, seq, list(qual)) == (headerw, seqw, list(qualw))



from __future__ import division
from builtins import zip
import pytest
import numpy
import itertools
import pandas
from collections import Counter
from mgkit.utils import trans_tables
from mgkit.utils.sequence import reverse_aa_coord, get_variant_sequence, \
    convert_aa_to_nuc_coord, reverse_complement, translate_sequence, \
    put_gaps_in_nuc_seq, get_seq_expected_syn_count, Alignment, \
    get_seq_number_of_syn, sequence_gc_ratio, sequence_gc_content, \
    sequence_composition, _get_kmers, _sliding_window, _sequence_signature, \
    sequence_signature, sliding_window, get_kmers, random_sequences_codon, \
    random_sequences, calc_n50


@pytest.mark.parametrize(
    "coords,result",
    [
        ((6, 8, 8), (1, 3)),
        ((2, 4, 8), (5, 7)),
        ((3, 5, 8), (4, 6)),
        ((1, 5, 8), (4, 8)),
        ((5, 8, 8), (1, 4)),
        ((10, 30, 100), (71, 91))
    ]
)
def test_reverse_aa_coord(coords, result):
    assert reverse_aa_coord(*coords) == result


@pytest.mark.parametrize(
    "seq,snps,result",
    [
        ['ACTGATATATGCGCGCATCT', [(1, 'C')], 'CCTGATATATGCGCGCATCT'],
        ['ACTGATATATGCGCGCATCT', [(1, 'C'), (7, 'G'), (5, 'N')], 'CCTGNTGTATGCGCGCATCT'],
        ['ACTGATATATGCGCGCATCT', [(1, 'C'), (7, 'G'), (5, 'N')], 'CCTGNTGTATGCGCGCATCT']
    ]
)
def test_get_variant_sequence(seq, snps, result):
    assert get_variant_sequence(seq, *snps) == result


@pytest.mark.parametrize(
    "coords,result",
    [
        [(1, 4, 0), (1, 12)],
        [(1, 4, 1), (2, 13)],
        [(3, 12, 0), (7, 36)],
        [(2, 23, 2), (6, 71)],
    ]
)
def test_convert_aa_to_nuc_coord(coords, result):
    assert convert_aa_to_nuc_coord(*coords) == result


def test_reverse_complement():
    seq = 'ACTGATATATGCGCGCATCTC'
    rev = 'GAGATGCGCGCATATATCAGT'

    assert reverse_complement(seq) == rev


@pytest.mark.parametrize(
    "seq,trl,start,tbl,reverse",
    [
        ('TTTAAAACCGGGGTC', 'FKTGV', 0, trans_tables.UNIVERSAL, False),
        ('TTTAAAACCGGGGTC', 'FKTGV', 0, None, False),
        ('TTTAAAACCGGGGTC', 'FKTGV', 0, trans_tables.VT_MIT, False),
        ('TTTAAAACCGGGGTF', 'FKTGX', 0, trans_tables.UNIVERSAL, False),
        ('TTTAAAACCGGGGTC', 'DPGFK', 0, trans_tables.UNIVERSAL, True),
    ]
)
def test_translate_sequence_eq(seq, trl, start, tbl, reverse):
    assert translate_sequence(seq, start=start, tbl=tbl, reverse=reverse) == trl

@pytest.mark.parametrize(
    "seq,trl,start,tbl,reverse",
    [
        ('TTTAAAACCGGGGTC', 'FKTGV', 1, trans_tables.UNIVERSAL, False),
        ('TTTAAAACCGGGGTC', 'FKTGV', 2, trans_tables.UNIVERSAL, False),
    ]
)
def test_translate_sequence_ne(seq, trl, start, tbl, reverse):
    assert translate_sequence(seq, start=start, tbl=tbl, reverse=reverse) != trl


@pytest.mark.parametrize(
    "nuc,aa_gap,nuc_gap, trim",
    [
        ('TTTAAAACCGGGGTC', 'FK-TG-V', 'TTTAAA---ACCGGG---GTC', True),
        ('TTTAAAACCGGGGTC', '-FKTG-V', '---TTTAAAACCGGG---GTC', True),
        ('TTTAAAACCGGGGTC', 'FKTG-V-', 'TTTAAAACCGGG---GTC---', True),
        ('TTTAAAACCGGGGTC', '-FKTG-V-', '---TTTAAAACCGGG---GTC---', True),
        # trim sequence by default
        ('TTTAAAACCGGGGTCAA', '-FKTG-V-', '---TTTAAAACCGGG---GTC---', True),
        # don't trim sequence by default
        ('TTTAAAACCGGGGTCAA', '-FKTG-V-', '---TTTAAAACCGGG---GTC---AA', False),
    ]
)
def test_put_gaps_in_nuc_seq(nuc, aa_gap, nuc_gap, trim):
    assert put_gaps_in_nuc_seq(nuc, aa_gap, trim) == nuc_gap


def test_get_seq_expected_syn_count():
    seq = 'ATGCATCGACTCTGCACTACG'

    assert get_seq_expected_syn_count(seq) == (15, 48)


@pytest.fixture
def alignment():
    return Alignment(
        seqs=[
            (1, 'ACTCACTCA'),
            (2, 'ACCCACCCT'),
            (3, 'ACCCGCCCT')
        ]
    )


def test_alignment1(alignment):
    assert alignment.get_position(6) == 'TCC'


def test_alignment2(alignment):
    assert alignment.get_seq_len() == 9


def test_alignment3(alignment):
    assert alignment.get_consensus() == 'ACCCACCCT'


def test_alignment4(alignment):
    ref_seq = alignment.get_consensus()
    snps = [(2, 'T'), (4, 'G'), (6, 'T'), (8, 'A')]
    assert alignment.get_snps(ref_seq) == snps


def test_alignment5(alignment):
    ref_seq = alignment.get_consensus()
    snps = [
        (0, None),
        (1, None),
        (2, 'T'),
        (3, None),
        (4, 'G'),
        (5, None),
        (6, 'T'),
        (7, None),
        (8, 'A')
    ]
    assert alignment.get_snps(ref_seq, full_size=True) == snps


@pytest.mark.parametrize(
    "ref_seq,snps,start,trans_table,result",
    [
        ('ACCCACCCT', [(2, 'T'), (4, 'G'), (6, 'T'), (8, 'A')], 0, trans_tables.UNIVERSAL, (2, 2)),
        ('ACCCACCCT', [(2, 'T'), (4, 'G'), (6, 'T'), (8, 'A')], 1, trans_tables.UNIVERSAL, (1, 2)),
        ('ACCCACCCT', [(2, 'T'), (4, 'G'), (6, 'T'), (8, 'A')], 2, trans_tables.UNIVERSAL, (1, 2)),
    ]
)
def test_get_seq_number_of_syn(ref_seq, snps, start, trans_table, result):
    assert get_seq_number_of_syn(ref_seq, snps, start=start, trans_table=trans_table) == result


def test_sequence_gc_ratio1():
    seq = 'A' * 10 + 'C' * 4 + 'T' * 4 + 'G' * 11
    assert sequence_gc_ratio(seq) == 14 / 15


def test_sequence_gc_ratio_fail1():
    seq = 'A' * 10 + 'T' * 4
    assert numpy.isnan(sequence_gc_ratio(seq))


def test_sequence_gc_content1():
    seq = 'A' * 10 + 'C' * 4 + 'T' * 4 + 'G' * 11
    sequence_gc_content(seq) == 15 / 29


def test_sequence_gc_content2():
    seq = ''
    sequence_gc_content(seq) == .5


def test_sequence_composition1():
    seq = 'A' * 10 + 'C' * 4 + 'T' * 5 + 'G' * 11
    assert sorted(sequence_composition(seq)) == [('A', 10), ('C', 4), ('G', 11), ('T', 5)]


def test_sequence_composition2():
    seq = 'A' * 10 + 'C' * 4 + 'T' * 5 + 'G' * 11 + 'N' * 2
    assert sorted(sequence_composition(seq, chars=None)) == [('A', 10), ('C', 4), ('G', 11), ('N', 2), ('T', 5)]


def test_sequence_composition3():
    seq = 'A' * 10 + 'C' * 4 + 'T' * 5 + 'G' * 11 + 'N' * 2
    assert sorted(sequence_composition(seq, chars=('A', 'C'))) == \
        [('A', 10), ('C', 4)]


def test_get_kmers1():
    seq = 'ACTG' * 2
    assert list(_get_kmers(seq, 4)) == ['ACTG', 'CTGA', 'TGAC', 'GACT', 'ACTG']


def test_get_kmers2():
    seq = 'ACTG' * 2
    assert list(_get_kmers(seq, 5)) == ['ACTGA', 'CTGAC', 'TGACT', 'GACTG']


def test_sliding_window1():
    seq = 'ACTG' * 5
    assert list(_sliding_window(seq, 4, 4)) == ['ACTG'] * 5


def test_sliding_window2():
    seq = 'ACTG' * 2
    assert list(_sliding_window(seq, 4, 2)) == ['ACTG', 'TGAC', 'ACTG']


def test_sliding_window3():
    seq = 'ACTG' * 2
    assert list(_sliding_window(seq, 4, 3)) == ['ACTG', 'GACT']


def test_sequence_signature1():
    seq = 'ACTG' * 2
    count = _sequence_signature(seq, 4, 4, 4)
    assert len(count) == 2


def test_sequence_signature2():
    seq = 'ACTG' * 2
    count = _sequence_signature(seq, 4, 4, 4)
    assert count[0]['ACTG'] == 1


def test_sequence_signature3():
    seq = 'ACTG' * 2
    count = _sequence_signature(seq, 8, 4, 4)
    assert count[0]['ACTG'] == 2


def test_sequence_signature4():
    seq = 'ACTG' * 2
    count = _sequence_signature(seq, 5, 4, 5)
    assert len(count) == 1


def test_sequence_signature_cython():
    seq = 'ACTG' * 2
    countP = _sequence_signature(seq, 4, 4, 4)
    countC = sequence_signature(seq, 4, 4, 4)
    assert len(countP) == len(countC)


def test_sliding_window_cython():
    seq = 'ACTG' * 5
    assert list(
         _sliding_window(seq, 4, 4)
    ) == list(
         sliding_window(seq, 4, 4)
    )


def test_get_kmers_cython():
    seq = 'ACTG' * 2
    assert list(_get_kmers(seq, 4)) == list(get_kmers(seq, 4))


def test_random_sequences_codon1():
    seqs = random_sequences_codon(n=20, length=150, frame=0, codons=list(trans_tables.UNIVERSAL))
    it = itertools.chain(
        *(
            _sliding_window(seq, 3, 3)
            for seq in seqs
        )
    )
    p = pandas.Series(Counter(it))
    p = p.mean() / p.sum()

    exp_p = 1. / len(trans_tables.UNIVERSAL)
    assert p == exp_p


def test_random_sequences_codon2():
    seqs = random_sequences_codon(n=20, length=150, frame=0, codons=list(trans_tables.VT_MIT))
    it = itertools.chain(
        *(
            _sliding_window(seq, 3, 3)
            for seq in seqs
        )
    )
    p = pandas.Series(Counter(it))
    p = p.mean() / p.sum()

    exp_p = 1. / len(trans_tables.VT_MIT)
    assert p == exp_p


def test_random_sequences_codon3():
    exp_p = pandas.Series(
        numpy.random.normal(
            size=len(trans_tables.UNIVERSAL), loc=100, scale=50
        ),
        index=list(trans_tables.UNIVERSAL.keys())
    ).apply(abs)
    exp_p = (exp_p / exp_p.sum()).sort_index()

    seqs = random_sequences_codon(n=10**5, length=150, frame=0,
                                  codons=sorted(trans_tables.UNIVERSAL),
                                  p=exp_p.tolist())
    it = itertools.chain(
        *(
            _sliding_window(seq, 3, 3)
            for seq in seqs
        )
    )
    p = pandas.Series(Counter(it), index=exp_p.index)
    p = p / p.sum()

    assert all(p < (exp_p * 1.1)) and all(p > (exp_p * .9))


def test_calc_n50():
    assert calc_n50(numpy.array([100] * 10 + [200]* 5 + [500] * 3)) == 200



"""
Utility functions for the network package
"""
import requests


def url_open(url, data=None, headers=None, agent=None, get=True, stream=False):
    """
    .. versionchanged:: 0.3.4
        now uses *requests*

    Arguments:
        url (str): url to request
        data (dict): parameters to pass to the request
        headers (dict): any additional headers
        agent (str): user agent to use
        get (bool): True if the request is a GET, False for POST
        stream (bool): returns an iterator to stream over

    :param str url: url to request
    :param str data: data to add to the request
    :param bool compress: if the response should be compressed
    :param str agent: if supplied, the 'User-Agent' header we'll be added to
        the request
    :return: the response handle
    """
    if (agent is not None) and (headers is None):
        headers = {'user-agent': agent}
    if get:
        request = requests.get(url, params=data, headers=headers, stream=stream)
    else:
        request = requests.post(url, params=data, headers=headers, stream=stream)

    if stream:
        return request.iter_lines()
    else:
        return request


def url_read(url, data=None, agent=None, headers=None, get=True):
    """
    .. versionchanged:: 0.3.4
        now uses *requests*, removed *compressed* and added *headers*, *get*

    Opens an URL and reads the

    Wrapper of :func:`url_open` which reads the full response

    Arguments:
        url (str): url to request
        data (dict or None): data to add to the request
        headers (dict or None): additional headers
        agent (str or None): if supplied, the 'User-Agent' header we'll be
            added to the request
        get (bool): uses a GET operation if True, POST if False

    Returns:
        str: the response data
    """
    request = url_open(url, data=data, agent=agent, headers=headers, get=get, stream=False)
    return request.text



"""
.. versionadded:: 0.1.14

Code related to boxplots
"""
from __future__ import division
from builtins import zip
import logging
import numpy
from ..utils.common import deprecated
from .colors import float_to_hex_color

LOG = logging.getLogger(__name__)

from matplotlib.patches import Polygon

try:
    import seaborn as sns
except ImportError:
    sns = None

DEFAULT_BOXPLOT_FONTCONF = {
    'rotation': 'vertical',
    'fontsize': 8
}

DEFAULT_BOXPLOT_COLOURS = {
    'boxes': '#636363',
    'medians': '#f0f0f0',
    'whiskers': '#636363',
    'caps': 'black',
    'fliers': '#636363',
    'vals': '#636363',
}


def boxplot_dataframe_multindex(dataframe, axes, plot_order=None,
                                label_map=None, fonts=None, fill_box=True,
                                colours=None, data_colours=None,
                                box_vert=True):
    """
    .. versionadded:: 0.1.13

    .. todo::

        documentation

    The function draws a series of boxplots from a DataFrame object, whose
    order is directed by the iterable plot_order. The columns of each DataFrame
    row contains the values for each boxplot. An axes object is needed.

    :param dataframe: dataframe to plot
    :param iterable plot_order: row order used to plot the boxes
    :param axes: an axes instance
    :param dict label_map: a map that converts the items in plot_order to a
        label used on the plot X axes
    :param dict fonts: dictionary with properties for x axis labels,
        :data:`DEFAULT_BOXPLOT_FONTCONF` is used by default
    :param bool fill_box: if True each box is filled with the same colour of
        its outline
    :param dict colours: dictionary with properties for each boxplot if
        data_colours is None, whi overrides box, whiskers and fliers. Defaults
        to :data:`DEFAULT_BOXPLOT_COLOURS`
    :param dict data_colours: dictionary of colours for each boxplot, a set of
        colours can be obtained using func:`map_taxon_to_colours`

    :return: the plot data same as matplotlib boxplot function
    """

    if colours is not None:
        colours = dict(
            (feature, colours[feature]) if feature in colours else (feature, colour)
            for feature, colour in DEFAULT_BOXPLOT_COLOURS.items()
        )
        DEFAULT_BOXPLOT_COLOURS.copy().update(colours)
    else:
        colours = DEFAULT_BOXPLOT_COLOURS.copy()

    if fonts is not None:
        fonts = dict(
            (feature, fonts[feature]) if feature in fonts else (feature, option)
            for feature, option in DEFAULT_BOXPLOT_FONTCONF.items()
        )
        DEFAULT_BOXPLOT_FONTCONF.copy().update(fonts)
    else:
        fonts = DEFAULT_BOXPLOT_FONTCONF.copy()

    categories = set(dataframe.index.get_level_values(1))

    if (data_colours is None) and (sns is not None):
        data_colours = dict(
            zip(
                categories,
                sns.color_palette("hls", len(categories))
            )
        )

    if plot_order is None:
        plot_order = dataframe.index

    if label_map is None:
        label_map = []
        for label in dataframe.index.get_level_values(0):
            if label in label_map:
                continue
            label_map.append(label)

    plot_data = axes.boxplot(
        [dataframe.loc[x].dropna() for x in plot_order],
        vert=box_vert
    )

    for idx, row_id in enumerate(plot_order):
        category = row_id[1]
        box = plot_data['boxes'][idx]
        box.set_color(
            data_colours[category] if data_colours else colours['boxes']
        )
        if fill_box:
            box_coord = list(zip(box.get_xdata(), box.get_ydata()))
            polygon = Polygon(
                box_coord,
                facecolor=data_colours[category] if data_colours else colours['boxes']
            )
            axes.add_patch(polygon)

        plot_data['medians'][idx].set_color(colours['medians'])

    # It's got a different length (double the size of plot_order)
    for idx, tx in enumerate(plot_data['whiskers']):
        whisker = plot_data['whiskers'][idx]
        whisker.set_color(
            # data_colours[tx] if data_colours else colours['whiskers']
            colours['whiskers']
        )
        plot_data['caps'][idx].set_color(colours['caps'])

    for flier in plot_data['fliers']:
        flier.set_markerfacecolor(
            colours['fliers']
            # data_colours[tx] if data_colours else colours['fliers']
        )

    if box_vert:
        ltick_setfunc = axes.set_xticklabels
        vtick_getfunc = axes.get_yticklabels
        ptick_setfunc = axes.set_xticks
    else:
        ltick_setfunc = axes.set_yticklabels
        vtick_getfunc = axes.get_xticklabels
        ptick_setfunc = axes.set_yticks

    if fonts is not None:
        ptick_setfunc(
            numpy.arange(
                numpy.arange(1, len(categories) + 1).mean(),
                len(dataframe.index),
                len(categories)
            )
        )
        ltick_setfunc(
            label_map,
            rotation=fonts['rotation'],
            fontsize=fonts['fontsize']
        )
        for label in vtick_getfunc():
            label.set_fontsize(fonts['fontsize'])
    else:
        ltick_setfunc([])

    return plot_data


def add_values_to_boxplot(dataframe, ax, plot_data, plot_order,
                          data_colours=None, alpha=0.5, s=80, marker='o',
                          linewidth=0.01, box_vert=False):
    """
    .. versionadded:: 0.1.13

    .. versionchanged:: 0.1.14
        added *box_vert* parameter

    .. versionchanged:: 0.1.16
        changed default value for *linewidth*

    Adds the values of a dataframe used in :func:`boxplot_dataframe` to the
    plot. *linewidth* must be higher than 0 if a marker like *|* is used.

    A list of markers is available at
    `this page <http://matplotlib.org/api/markers_api.html>`_

    .. warning::

        Contrary to :func:`boxplot_dataframe`, the boxplot default is
        horizontal (*box_vert*). The default will change in a later version.

    Arguments:
        dataframe: dataframe with the values to plot
        ax: an axis instance
        plot_data: return value from :func:`boxplot_dataframe`
        plot_order (iterable): row order used to plot the boxes
        data_colours (dict): colors used for the values
        alpha (float): alpha value for the colour
        s (int): size of the marker drawn
        marker (str): one of the accepted matplotlib markers
        linewidth (float): width of the line used to draw the marker
        box_vert (bool): specify if the original boxplot is vertical or not
    """

    if data_colours is not None:
        # in case each color is a tuple of rgb floats, it converts them into
        # strings to avoid matplotlit to confuse the single color as different
        # shades of grey. It only happens when the number of data points in a
        # row is 3
        if not isinstance(data_colours[list(data_colours.keys())[0]], str):
            data_colours = dict(
                (key, float_to_hex_color(*value))
                for key, value in data_colours.items()
            )

    for index, row_id in enumerate(plot_order):
        if box_vert:
            xvals = plot_data['medians'][index].get_xdata()
            mean_x = xvals.mean()
            y = dataframe.loc[row_id].dropna()
            x = [mean_x] * dataframe.loc[row_id].count()
        else:
            yvals = plot_data['medians'][index].get_ydata()
            mean_y = yvals.mean()
            x = dataframe.loc[row_id].dropna()
            y = [mean_y] * dataframe.loc[row_id].count()

        if data_colours is None:
            DEFAULT_BOXPLOT_COLOURS['boxes']
        else:
            color = data_colours[row_id]

        ax.scatter(
            x,
            y,
            c=color,
            alpha=alpha,
            s=s,
            marker=marker,
            linewidth=linewidth,
            # this option put the dots below the lines of the boxplot
            zorder=1
        )


def boxplot_dataframe(dataframe, plot_order, ax, label_map=None, fonts=None,
                      fill_box=True, colours=None, data_colours=None,
                      box_vert=True, widths=0.5):
    """
    .. versionadded:: 0.1.7
        To move from an all-in-one drawing to a more modular one.

    .. versionchanged:: 0.1.13
        added box_vert parameter

    .. versionchanged:: 0.1.16
        added *widths* parameter

    The function draws a series of boxplots from a DataFrame object, whose
    order is directed by the iterable plot_order. The columns of each DataFrame
    row contains the values for each boxplot. An ax object is needed.

    :param dataframe: dataframe to plot
    :param iterable plot_order: row order used to plot the boxes
    :param ax: an axis instance
    :param dict label_map: a map that converts the items in plot_order to a
        label used on the plot X ax
    :param dict fonts: dictionary with properties for x axis labels,
        :data:`DEFAULT_BOXPLOT_FONTCONF` is used by default
    :param bool fill_box: if True each box is filled with the same colour of
        its outline
    :param dict colours: dictionary with properties for each boxplot if
        data_colours is None, whi overrides box, whiskers and fliers. Defaults
        to :data:`DEFAULT_BOXPLOT_COLOURS`
    :param dict data_colours: dictionary of colours for each boxplot, a set of
        colours can be obtained using func:`map_taxon_to_colours`
    :param bool box_vert: if False the boxplots are drawn horizontally
    :param float widths: width (scalar or array) of the boxplots width(s)

    :return: the plot data; same as matplotlib boxplot function
    """

    if colours is not None:
        colours = dict(
            (feature, colours[feature]) if feature in colours else (feature, colour)
            for feature, colour in DEFAULT_BOXPLOT_COLOURS.items()
        )
        # DEFAULT_BOXPLOT_COLOURS.copy().update(colours)
    else:
        colours = DEFAULT_BOXPLOT_COLOURS.copy()

    if fonts is not None:
        fonts = dict(
            (feature, fonts[feature]) if feature in fonts else (feature, option)
            for feature, option in DEFAULT_BOXPLOT_FONTCONF.items()
        )
        DEFAULT_BOXPLOT_FONTCONF.copy().update(fonts)
    else:
        fonts = DEFAULT_BOXPLOT_FONTCONF.copy()

    plot_data = ax.boxplot(
        [dataframe.loc[x].dropna() for x in plot_order],
        vert=box_vert,
        widths=widths
    )

    for idx, row_id in enumerate(plot_order):
        box = plot_data['boxes'][idx]
        box.set_color(
            data_colours[row_id] if data_colours else colours['boxes']
        )
        if fill_box:
            box_coord = list(zip(box.get_xdata(), box.get_ydata()))
            polygon = Polygon(
                box_coord,
                facecolor=data_colours[row_id] if data_colours else colours['boxes']
            )
            ax.add_patch(polygon)

        plot_data['medians'][idx].set_color(colours['medians'])

    # It's got a different length (double the size of plot_order)
    for idx, tx in enumerate(plot_data['whiskers']):
        whisker = plot_data['whiskers'][idx]
        whisker.set_color(
            # data_colours[tx] if data_colours else colours['whiskers']
            colours['whiskers']
        )
        plot_data['caps'][idx].set_color(colours['caps'])

    for flier in plot_data['fliers']:
        flier.set_color(
            colours['whiskers']
            # data_colours[tx] if data_colours else colours['fliers']
        )

    if box_vert:
        ltick_setfunc = ax.set_xticklabels
        vtick_getfunc = ax.get_yticklabels
    else:
        ltick_setfunc = ax.set_yticklabels
        vtick_getfunc = ax.get_xticklabels

    if fonts is not None:
        ltick_setfunc(
            [
                label if label_map is None else label_map[label]
                for label in plot_order
            ],
            rotation=fonts['rotation'], fontsize=fonts['fontsize']
        )
        for label in vtick_getfunc():
            label.set_fontsize(fonts['fontsize'])
    else:
        ltick_setfunc([])

    return plot_data


def add_significance_to_boxplot(sign_indices, ax, pos, box_vert=True,
                                fontsize=16):
    """
    .. versionadded:: 0.1.16

    Add significance groups to boxplots

    Arguments:
        sign_indices (iterable): iterable in which each element is a tuple;
            each element of the tuple is the numerical index of the position of
            the significant boxplot
        ax: an axis instance
        pos (tuple): the 2 values are the coordinates for the top line, and the
            the lowest bound for the whisker
        box_vert (bool): if the boxplot is vertical
        fontsize (float): size for the * (star)
    """
    maxval, spine = pos
    text = maxval

    for index, (sign1, sign2) in enumerate(sign_indices):
        spacer = (index * ((maxval - spine) * 2.5))
        factors = [spine, maxval]
        x = numpy.array(
            factors + factors[::-1]
        ) - spacer

        y = [sign1 + 1, sign1 + 1, sign2 + 1, sign2 + 1]

        if box_vert:
            x, y = y, x
        ax.plot(x, y, linestyle='-', c='k', alpha=.75)
        if box_vert:
            xtext, ytext = numpy.mean(x), text - spacer
        else:
            xtext, ytext = text - spacer, numpy.mean(y)
        ax.text(xtext, ytext, '*', fontsize=fontsize)

__all__ = [
    'add_values_to_boxplot',
    'add_significance_to_boxplot',
    'boxplot_dataframe_multindex',
    'boxplot_dataframe'
]



import pytest
from mgkit.io.gff import Annotation, from_gff
from mgkit.db.dbm import create_gff_dbm, GFFDB

@pytest.fixture
def annotations():
    return [
        Annotation(seq_id='1', start=0, end=100, strand='+', uid='u1'),
        Annotation(seq_id='1', start=10, end=100, strand='-', uid='u2'),
        Annotation(seq_id='2', start=0, end=100, strand='+', uid='u3'),
    ]

@pytest.fixture
def dbm_file(annotations, tmpdir):
    return create_gff_dbm(annotations, tmpdir.join('test.db').strpath)


@pytest.fixture
def gff_db(dbm_file):
    return GFFDB(dbm_file)


def test_create_gff_dbm(dbm_file):
    assert dbm_file


def test_create_gff_data1(annotations, dbm_file):
    assert from_gff(dbm_file['u1']) == annotations[0]


def test_create_gff_data2(annotations, dbm_file):
    assert from_gff(dbm_file['u2']) == annotations[1]


def test_create_gff_data3(annotations, dbm_file):
    assert from_gff(dbm_file['u3']) == annotations[2]


def test_gffdb_getitem(gff_db, annotations):
    assert gff_db['u1'] == annotations[0]


def test_gff_db_init_path(tmpdir, annotations):
    db = GFFDB(tmpdir.join('test2.db').strpath)
    db['u1'] = annotations[0]
    assert db['u1'] == annotations[0]


def test_gffdb_iter(gff_db, annotations):
    assert set(x.uid for x in annotations) == set(x for x in gff_db)


def test_gffdb_values(gff_db, annotations):
    assert set(x.uid for x in annotations) == set(x.uid for x in gff_db.values())


def test_gffdb_itervalues(gff_db, annotations):
    assert set(x.uid for x in annotations) == set(x.uid for x in gff_db.itervalues())


def test_gffdb_items(gff_db, annotations):
    assert set(x.uid for x in annotations) == set(x[0] for x in gff_db.items())


def test_gffdb_iteritems(gff_db, annotations):
    assert set(x.uid for x in annotations) == set(x[0] for x in gff_db.iteritems())



"""
Contains function and constants for Uniprot access
"""
from __future__ import division
from builtins import range, zip
import mgkit
import logging
from . import url_read

UNIPROT_MAP = 'http://www.uniprot.org/mapping/'
"URL to Uniprot mapping REST API"

UNIPROT_GET = 'http://www.uniprot.org/uniprot/'
"URL to Uniprot REST API"

UNIPROT_TAXONOMY = 'http://www.uniprot.org/taxonomy/'
"URL to Uniprot REST API - Taxonomy"

COLS_TAXON = 'organism-id'
COLS_KO = 'database(KO)'
COLS_EGGNOG = 'database(EGGNOG)'
COLS_EC = 'ec'

LOG = logging.getLogger(__name__)


def get_sequences_by_ko(ko_id, taxonomy, contact=None, reviewed=True):
    """
    Gets sequences from Uniprot, restricting to the taxon id passed.

    :param str ko_id: KO id of the sequences to download
    :param int taxonomy: id of the taxon
    :param str contact: email address to be passed in the query (requested by
        Uniprot API)
    :param bool reviewed: if the sequences requested must be reviewed

    :return: string with the fasta file downloaded
    """
    params = {
        'query': 'database:(type:ko {0}) AND taxonomy:{1}{2}'.format(
            ko_id, taxonomy, ' reviewed:yes' if reviewed else ''),
        'format': 'fasta',
        'limit': 200,
        'sort': 'score'
    }
    if mgkit.DEBUG:
        LOG.debug("query: %s?%s", UNIPROT_GET, params)
        LOG.debug("request length %d", len(params))

    fasta = url_read(UNIPROT_GET, data=params, agent=contact)

    return fasta


def get_mappings(entry_ids, db_from='ID', db_to='EMBL', out_format='tab',
                 contact=None):
    """
    Gets mapping of genes using Uniprot REST API. The db_from and db_to values
    are the ones accepted by Uniprot API. The same applies to out_format, the
    only processed formats are 'list', which returns a list of the mappings
    (should be used with one gene only) and 'tab', which returns a dictionary
    with the mapping. All other values returns a string with the newline
    stripped.

    :param iterable entry_ids: iterable of ids to be mapped (there's a limit)
        to the maximum length of a HTTP request, so it should be less than 50
    :param str db_from: string that identify the DB for elements in entry_ids
    :param str db_to: string that identify the DB to which map entry_ids
    :param str out_format: format of the mapping; 'list' and 'tab' are
        processed
    :param str contact: email address to be passed in the query (requested
        Uniprot API)

    :return: tuple, dict or str depending on out_format value
    """

    if isinstance(entry_ids, str):
        entry_ids = [entry_ids]

    data = {
        'from': db_from,
        'to': db_to,
        'query': ' '.join(entry_ids),
        'format': out_format
    }

    mappings = url_read(UNIPROT_MAP, data=data, agent=contact)

    mappings = mappings.strip()

    if out_format == 'list':
        mappings = mappings.split('\n')
    elif out_format == 'tab':
        mapping_dict = {}
        mappings = mappings.split('\n')
        # delete first row 'From to'
        del mappings[0]

        for mapping in mappings:
            id_from, id_to = mapping.split('\t')
            if id_to == 'null':
                continue
            try:
                mapping_dict[id_from].append(id_to)
            except KeyError:
                mapping_dict[id_from] = [id_to]

        mappings = mapping_dict

    return mappings


def ko_to_mapping(ko_id, query, columns, contact=None):
    """
    Returns the mappings to the supplied KO. Can be used for any id, the
    query format is free as well as the columns returned. The only
    restriction is using a tab format, that is parsed.

    :param str ko_id: id used in the query
    :param str query: query passed to the Uniprot API, ko_id is replaced
        using :func:`str.format`
    :param str column: column used in the results table used to map the ids
    :param str contact: email address to be passed in the query (requested
        Uniprot API)

    .. note::

        each mapping in the column is separated by a ;

    """
    data = {
        'query': query.format(ko_id),
        'format': 'tab',
        'columns': columns
    }

    mappings = url_read(UNIPROT_GET, data=data, agent=contact)

    if mgkit.DEBUG:
        LOG.debug("query: %s?%s", UNIPROT_GET, data)
        LOG.debug("request length %d", len(data))

    mappings = mappings.split('\n')
    del mappings[0]

    categories = set()

    for map_line in mappings:

        mappings = [mapping.strip() for mapping in map_line.split(';')]
        if not mappings:
            continue

        categories.update(mappings)

    # in case an empty line is present
    try:
        categories.remove('')
    except KeyError:
        pass

    return categories


def get_gene_info(gene_ids, columns, max_req=50, contact=None):
    """
    .. versionadded:: 0.1.12

    Get informations about a list of genes. it uses :func:`query_uniprot` to
    send the request and format the response in a dictionary.

    Arguments:
        gene_ids (iterable, str): gene id(s) to get informations for
        columns (list): list of columns
        max_req (int): number of maximum *gene_ids* per request
        contact (str): email address to be passed in the query (requested
            Uniprot API)

    Returns:
        dict: dictionary where the keys are the *gene_ids* requested and the
        values are dictionaries with the names of the *columns* requested as
        keys and the corresponding values, which can be lists if the values are
        are semicolon separated strings.

    Example:
        To get the taxonomy ids for some genes:

        >>> uniprot.get_gene_info(['Q09575', 'Q8DQI6'], ['organism-id'])
        {'Q09575': {'organism-id': '6239'}, 'Q8DQI6': {'organism-id': '171101'}}

    """
    if isinstance(gene_ids, str):
        gene_ids = [gene_ids]
    elif isinstance(gene_ids, set):
        gene_ids = list(gene_ids)

    if isinstance(columns, str):
        columns = [columns]

    infos = {}

    for index in range(0, len(gene_ids), max_req):

        LOG.info(
            "Querying uniprot ids (%d/%d)",
            index + max_req,
            len(gene_ids)
        )

        info_lines = query_uniprot(
            ' OR '.join('id:{}'.format(gene_id) for gene_id in gene_ids[index:index+max_req]),
            columns=['id'] + columns,
            contact=contact
        )

        info_lines = info_lines.split('\n')

        del info_lines[0]

        for info_line in info_lines:
            info_line = info_line.strip()
            if not info_line:
                continue
            values = info_line.split('\t')

            gene_id = values[0]

            infos[gene_id] = dict(
                (
                    column,
                    value if (not value.endswith(';')) and (not value.endswith('; ')) and ('; ' not in value)
                    else [x.strip() for x in value.split(';') if x.strip()]
                )
                for column, value in zip(columns, values[1:])
            )

    return infos


def query_uniprot(query, columns=None, format='tab', limit=None, contact=None,
                  baseurl=UNIPROT_GET):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.1.13
        added *baseurl* and made *columns* a default argument

    Queries Uniprot, returning the raw response in tbe format specified. More
    informations at the `page <http://www.uniprot.org/faq/28>`_

    Arguments:
        query (str): query to submit, as put in the input box
        columns (None, iterable): list of columns to return
        format (str): response format
        limit (int, None): number of entries to return or *None* to request all
            entries
        contact (str): email address to be passed in the query (requested
            Uniprot API)
        baseurl (str): base url for the REST API, can be either
            :data:`UNIPROT_GET` or :data:`UNIPROT_TAXONOMY`
    Returns:
        str: raw response from the query

    Example:
        To get the taxonomy ids for some genes:

        >>> uniprot.query_uniprot('Q09575 OR Q8DQI6', ['id', 'organism-id'])
        'Entry\\tOrganism ID\\nQ8DQI6\\t171101\\nQ09575\\t6239\\n'

    .. warning::

        because of limits in the length of URLs, it's advised to limit the
        length of the query string.

    """
    data = {
        'query': query,
        'format': format
    }

    if limit is not None:
        data['limit'] = limit

    if columns is not None:
        data['columns'] = ','.join(columns)

    if mgkit.DEBUG:
        LOG.debug("query: %s?%s", baseurl, data)
        LOG.debug("request length %d", len(data))

    return url_read(baseurl, data, agent=contact)


def parse_uniprot_response(data, simple=True):
    """
    .. versionadded:: 0.1.12

    Parses raw response from a Uniprot query (tab format only) from functions
    like :func:`query_uniprot` into a dictionary. It requires that the first
    column is the entry id (or any other unique id).

    Arguments:
        data (str): string response from Uniprot
        simple (bool): if True and the number of columns is 1, the dictionary
            returned has a simplified structure

    Returns:
        dict: The format of the resulting dictionary is
        entry_id -> {column1 -> value, column2 -> value, ..} unless there's
        only one column and *simple* is True, in which case the value is
        equal to the value of the only column.
    """
    data = data.splitlines()

    columns = [x.lower() for x in data[0].split('\t')[1:]]

    del data[0]

    parsed_data = {}

    for line in data:
        line = line.split('\t')
        entry_id = line[0]

        if (len(columns) == 1) and simple:
            parsed_data[entry_id] = line[1]
        else:
            parsed_data[entry_id] = dict(
                zip(columns, line[1:])
            )

    return parsed_data


def get_ko_to_eggnog_mappings(ko_ids, contact=None):
    """
    .. versionadded:: 0.1.14

    It's not possible to map in one go KO IDs to eggNOG IDs via the API in
    Uniprot. This function uses :func:`query_uniprot` to get all Uniprot IDs
    requested and the return a dictionary with all their eggNOG IDs they map
    to.

    Arguments:
        ko_ids (iterable): an iterable of KO IDs
        contact (str): email address to be passed in the query (requested
            Uniprot API)

    Returns:
        dict: The format of the resulting dictionary is
        ko_id -> {eggnog_id1, ..}
    """

    data = query_uniprot(
        "database:(type:ko AND ({}))".format(' OR '.join(list(ko_ids))),
        columns=['database(KO)', 'database(EGGNOG)'],
        contact=contact
    )

    data = data.splitlines()

    del data[0]

    parsed_data = {}

    for line in data:
        ko_ids, eggnog_ids = line.split('\t')
        ko_ids = ko_ids.split(';')

        for ko_id in ko_ids:
            if not ko_id:
                continue
            if not eggnog_ids:
                continue

            for eggnog_id in eggnog_ids.split(';'):
                if not eggnog_id:
                    continue

                try:
                    parsed_data[ko_id].add(eggnog_id)
                except KeyError:
                    parsed_data[ko_id] = set([eggnog_id])

    return parsed_data


def get_uniprot_ec_mappings(gene_ids, contact=None):
    """
    .. versionadded:: 0.1.14

    Shortcut to download EC mapping of Uniprot IDs. Uses :func:`get_gene_info`
    passing the correct column (*ec*).

    """
    return get_gene_info(
        gene_ids,
        columns=['ec'],
        contact=contact,
        max_req=100
    )


def get_gene_info_iter(gene_ids, columns, contact=None, max_req=50):
    """

    .. versionadded:: 0.3.3

    Alternative function to :func:`get_gene_info`, returning an iterator to
    avoid connections timeouts when updating a dictionary

    This funciton's parameters are the same as :func:`get_gene_info`
    """

    gene_ids = list(gene_ids)

    for index in range(0, len(gene_ids), max_req):
        yield get_gene_info(
            gene_ids[index:index+max_req],
            columns,
            contact=contact,
            max_req=max_req
        )



"""
SNPs filtering functions
"""
import functools
from builtins import filter
from .. import consts
from ..filter.taxon import filter_taxon_by_id_list
from ..filter.common import FilterFails


def filter_genesyn_by_taxon_id(gene_syn, taxonomy=None, filter_list=None,
                               exclude=False, func=None):
    """
    Checks if the `taxon_id` attribute of `gene_syn` is the `filter_list`.
    Excelude reverses the result. If func is supplied, it's used to traverse
    the `taxonomy`.

    Arguments:
        gene_syn: :class:`~mgkit.snps.GeneSyn` instance
        taxonomy: a valid taxonomy (instance of
            :class:`~mgkit.taxon.Taxonomy`)
        filter_list (iterable): list of taxon IDs to include/exclude
        exclude (bool): if the filter is reversed
        func (func): :func:`~mgkit.taxon.is_ancestor`

    Returns:
        bool: if the exclude is True, the gene_id must appear in the gene_ids,
        if False, returns True only if gene_id is NOT in gene_ids.

    Raises:
        FilterFails: if filter_list is None or taxonomy is None and func is not
            None

    """

    if ((taxonomy is None) and (func is not None)) or (filter_list is None):
        raise FilterFails('No taxonomy supllied')

    if func is not None:
        func = functools.partial(func, taxonomy)

    taxon_id = gene_syn.taxon_id

    return filter_taxon_by_id_list(
        taxon_id,
        filter_list=filter_list,
        exclude=exclude,
        func=func
    )


def filter_genesyn_by_gene_id(gene_syn, gene_ids=None, exclude=False,
                              id_func=None):
    """
    Checks if the gene_id is listed in the filter_list.

    Arguments:
        gene_syn: :class:`~mgkit.snps.GeneSyn` instance
        gene_ids (iterable): list of gene IDs to include/exclude
        exclude (bool): if the filter is reversed

    Returns:
        bool: if the exclude is True, the gene_id must appear in the gene_ids,
        if False, returns True only if gene_id is NOT in gene_ids.

    Raises:
        FilterFails: if gene_ids is None

    """
    if gene_ids is None:
        raise FilterFails('No gene_ids supplied')
    return (id_func(gene_syn) in gene_ids) ^ exclude


def filter_genesyn_by_coverage(gene_syn, min_cov=None):
    """
    Checks if the coverage of the provided `gene_syn` is at least `min_cov`

    Arguments:
        gene_syn: :class:`~mgkit.snps.GeneSyn` instance
        min_cov (int): minimum coverage allowed (included)

    Returns:
        bool: True if the gene has enough coverage

    Raises:
        FilterFails: if min_cov is None
    """
    if min_cov is None:
        raise FilterFails('No coverage supplied')

    return gene_syn.coverage >= min_cov


def get_default_filters(taxonomy, **kwargs):
    """
    Retuns a list of filters that are used by default. it needs a valid
    taxonomy and gets the default arguments from
    :data:`mgkit.consts.DEFAULT_SNP_FILTER`.
    """
    filter_opts = consts.DEFAULT_SNP_FILTER.copy()
    filter_opts.update(kwargs)

    filter_coverage = functools.partial(
        filter_genesyn_by_coverage,
        min_cov=filter_opts['min_cov']
    )
    filter_black_list = functools.partial(
        filter_genesyn_by_taxon_id,
        taxonomy=taxonomy,
        filter_list=filter_opts['include_only'],
        exclude=False,
        func=filter_opts['func']
    )
    return [filter_coverage, filter_black_list]


def pipe_filters(iterable, *funcs):
    """
    Pipes a list of filter to iterable, using the python ifilter function in
    the itertools module. Now using `builtins.filter`
    """
    for func in funcs:
        iterable = filter(func, iterable)
    for value in iterable:
        yield value



"""
Manage SNP data.
"""

from __future__ import division
from builtins import object
import logging
import enum
import json

from .. import DependencyError

try:
    import numpy
except ImportError:
    raise DependencyError('numpy')

LOG = logging.getLogger(__name__)


class RatioMixIn(object):
    def calc_ratio(self, haplotypes=False):
        """
        .. versionchanged:: 0.2.2
            split the function to handle *flag_value* in another method

        Calculate :math:`\\frac {pN}{pS}` for the gene.

        .. math::
            :label: pn-ps

            \\frac {pN}{pS} = \\frac{ ^{oN}/_{eN}}{ ^{oS}/_{eS}}

        WHere:

        * oN (number of non-synonymous - **nonsyn**)
        * eN (expected number of non-synonymous - **exp_nonsyn**)
        * oS (number of synonymous - **syn**)
        * eS (expected number of synonymous - **exp_syn**)

        Arguments:
            flag_value (bool): when there's no way to calculate the ratio, the
                possible cases will be flagged with a negative number. This
                allows to make substitutions for these values
            haplotypes (bool): if true, coverage information is not used,
                because the SNPs are assumed to come from an alignment that has
                sequences having haplotypes

        Returns:
            float: the :math:`\\frac {pN}{pS}` for the gene.

            .. note::

                Because pN or pS can be 0, and the return value would be NaN,
                we take in account some special cases. The default return value
                in this cases is :const:`numpy.nan`.

            * Both synonymous and non-synonymous values are 0:

                * if both the syn and nonsyn attributes are 0 but there's
                  coverage for this gene, we return a 0, as there's no
                  evolution in this gene. Before, the coverage was checked by
                  this method against either the passed *min_cov* parameter
                  that was equal to :data:`MIN_COV`. Now the case is for the
                  user to check the coverage and functions in
                  :mod:`mgkit.snps.conv_func` do that. If enough coverage was
                  achieved, the *haplotypes* parameter can be used to return a
                  0

            All other cases return a NaN value

        """

        # Both values are non-zero
        if (self.nonsyn != 0) and (self.syn != 0):
            pn_value = self.nonsyn / self.exp_nonsyn
            ps_value = self.syn / self.exp_syn
            return pn_value / ps_value
        # case in which a the SNPs come from haplotypes, in this case we don't
        # need to check for coverage to return a 0 for this special case
        elif (self.nonsyn == 0) and (self.syn == 0) and haplotypes:
            return 0

        return numpy.nan

    def calc_ratio_flag(self):
        """
        .. versionadded:: 0.2.2

        Handles cases where it's important to flag the returned value, as
        explained in :meth:`GeneSNP.calc_ratio`, and when the both the number
        of synonymous and non-synonymous is greater than 0, the pN/pS value is
        returned.

        * The number of non-synonymous is greater than 0 but the number of
              synonymous is 0:

                * if **flag_value** is **True**, the returned value is **-1**

            * The number of synonymous is greater than 0 but the number of
              non-synonymous is 0:

                * if **flag_value** is **True**, the returned value is **-2**

        +------------+------------+--------------+
        | :math:`oS` | :math:`oN` | return value |
        +============+============+==============+
        | >0         | >0         | **pN/pS**    |
        +------------+------------+--------------+
        | 0          | 0          | **-3**       |
        +------------+------------+--------------+
        | >0         | 0          | **-1**       |
        +------------+------------+--------------+
        | 0          | >0         | **-2**       |
        +------------+------------+--------------+

        """
        # Both values are non-zero
        if (self.nonsyn != 0) and (self.syn != 0):
            pn_value = self.nonsyn / self.exp_nonsyn
            ps_value = self.syn / self.exp_syn
            return pn_value / ps_value

        if self.nonsyn != 0:
            # there's at least non-synonymous count but no synonymous one
            # this will be converted in the max value for the matrix or
            # to some other value (-1 flag this case)
            if self.syn == 0:
                return -1
        else:
            # there's at least a synonymous count but no non-synonymous one
            # this will be converted in the max value for the matrix or
            # to some other value (-2 flag this case)
            if self.syn != 0:
                return -2
            else:
                # There's no changes in the gene at at all. It should be
                # checked if that gene has coverage.
                return -3

    def calc_ps(self):
        """
        Method that returns only the pS part of the pN/pS ratio.

        Returns:
            float: the pS value, unless self.syn is 0, in which case `numpy.nan` is returned
        """
        if self.syn == 0:
            return numpy.nan
        return self.syn / self.exp_syn


    def calc_pn(self):
        """
        Method that returns only the pN part of the pN/pS ratio.

        Returns:
            float: the pN value, unless self.nonsyn is 0, in which case `numpy.nan` is returned
        """
        if self.nonsyn == 0:
            return numpy.nan
        return self.nonsyn / self.exp_nonsyn


class SNPType(enum.Enum):
    """
    .. versionadded:: 0.1.13

    Enum that defines SNP types. Supported at the moment:

    * unknown = 0
    * syn (synonymous) = 1
    * nonsyn (non-synonymous) = 2

    .. note::

        No support is planned at the moment to support indel mutations

    """
    unknown = 0
    syn = 1
    nonsyn = 2


class GeneSNP(RatioMixIn):
    """
    .. versionadded:: 0.1.13

    Class defining gene and synonymous/non-synonymous SNPs.

    It defines background synonymous/non-synonymous attributes and only has a
    method right now, which calculate pN/pS ratio. The method is added through
    a mixin object, so the ratio can be customised and be shared with the old
    implementation.

    Attributes:
        uid (str): unique id for the isoform (to be referenced in a GFF file)
        gene_id (str): gene id
        taxon_id (int): gene taxon
        exp_syn (int): expected synonymous changes
        exp_nonsyn (int): expected non-synonymous changes
        coverage (int): gene coverage
        snps (list): list of SNPs associated with the gene, each element is a
            tuple with the position (relative to the gene start), the second is
            the nucleotidic change and the third is the aa SNP type as defined
            by :class:`SNPType`.

    .. note::

        The main difference with the :class:`GeneSyn` is that all snps are kept
        and `syn` and `nonsyn` are not attributes but properties that return
        the count of synonymous and non-synonymous SNPs in the `snps` list.

    .. warning::

        This class uses more memory than :class:`GeneSyn` because it doesn't
        use __slots__, it may be changed in later versions.

    """
    uid = None
    gene_id = None
    taxon_id = None
    exp_syn = None
    exp_nonsyn = None
    coverage = None
    snps = None

    def __init__(self, gene_id='', taxon_id=0, exp_syn=0, exp_nonsyn=0,
                 coverage=None, snps=None, uid=None, json_data=None):
        """
        If json_data is passed, that's used to initialise the instance
        """

        if json_data is not None:
            self.from_json(json_data)
            return

        self.uid = uid
        self.gene_id = gene_id
        self.taxon_id = taxon_id
        self.exp_syn = exp_syn
        self.exp_nonsyn = exp_nonsyn
        self.coverage = coverage
        self.snps = [] if snps is None else snps

    def add_snp(self, position, change, snp_type=SNPType.unknown):
        """
        Adds a SNP to the list

        Arguments:
            position (int): SNP position, relative to the gene start
            change (str): nucleotidic change
            snp_type (enum): one of the values defined in :class:`SNPType`
        """
        self.snps.append(
            (position, change, snp_type)
        )

    def add(self, other):
        """
        Inplace addition of another instance values. No check for them being
        the same gene/taxon, it's up to the user to check that they can be
        added together.

        Arguments:
            other: instance of :class:`GeneSyn` to add
        """

        self.snps.extend(other.snps)

        self.exp_syn += other.exp_syn
        self.exp_nonsyn += other.exp_nonsyn

        if other.coverage is not None:
            if self.coverage is None:
                self.coverage = other.coverage
            else:
                self.coverage += other.coverage

    def to_json(self):
        """
        Returns a json definition of the instance

        Returns:
            str: json representation of the instance
        """
        return json.dumps(
            {
                'uid': self.uid,
                'gene_id': self.gene_id,
                'exp_syn': self.exp_syn,
                'exp_nonsyn': self.exp_nonsyn,
                'taxon_id': self.taxon_id,
                'coverage': self.coverage,
                'snps': [
                    (pos, change, snptype.value)
                    for pos, change, snptype in self.snps
                ]
            }
        )

    def from_json(self, data):
        """
        Instantiate the instance with values from a json definition

        Arguments:
            data (str): json representation, as returned by
                :meth:`GeneSNP.to_json`
        """
        data = json.loads(data)
        self.uid = str(data['uid'])
        self.gene_id = str(data['gene_id'])
        self.exp_syn = data['exp_syn']
        self.exp_nonsyn = data['exp_nonsyn']
        self.taxon_id = data['taxon_id']
        self.coverage = data['coverage']
        self.snps = [
            (pos, str(change), SNPType(snptype))
            for pos, change, snptype in data['snps']
        ]

    def _cache_values(self):
        self._syn = sum(1 for x in self.snps if x[2] is SNPType.syn)
        self._nonsyn = sum(1 for x in self.snps if x[2] is SNPType.nonsyn)

    def _get_cached(self, attr):
        if (getattr(self, '_nsnps', None) is None) or \
                (self._nsnps != len(self.snps)):
            self._cache_values()
        return getattr(self, attr)

    @property
    def syn(self):
        "Returns the expected synonymous changes"
        return self._get_cached('_syn')

    @property
    def nonsyn(self):
        "Returns the expected non-synonymous changes"
        return self._get_cached('_nonsyn')



from mgkit.io import open_file


def parse_glimmer3(file_handle):
    """
    Parses an ouput file from glimmer3 and yields the header and prediction
    lines. Used to feed the :func:`mgkit.io.gff.from_glimmer3` function.

    Arguments:
        file_handle (str, file): file name or file handle to read from

    Yields:
        tuple: first element is the sequence of the predicted gene and the
        second is the prediction line
    """
    if isinstance(file_handle, str):
        file_handle = open_file(file_handle, 'r')

    curr_seq = ''
    predictions = []

    for line in file_handle:
        line = line.strip()
        if line.startswith('>'):
            if len(predictions) > 0:
                for prediction in predictions:
                    yield curr_seq, prediction
            curr_seq = line[1:]
            predictions = []
        else:
            if line != '':
                predictions.append(line)
    else:
        if len(predictions) > 0:
            for prediction in predictions:
                yield curr_seq, prediction



"""
Wappers to use some of the general function of the snps package
in a simpler way.
"""
import functools
import mgkit.snps.funcs
import mgkit.snps.filter
import mgkit.snps.mapper


def get_rank_dataframe(snp_data, taxonomy, min_num=3, rank='order',
                       index_type='taxon', filters=None):
    """
    .. versionadded:: 0.1.11

    .. versionchanged:: 0.2.2
        added *filters* argument

    Returns a :class:`~pandas.DataFrame` with the pN/pS of the given
    SNPs data, mapping all taxa to the specified rank. Higher taxa won't
    be included.

    Shortcut for using :func:`~mgkit.snps.funcs.combine_sample_snps`, using
    filters from :func:`~mgkit.snps.filter.get_default_filters` and as
    `taxon_func` parameter :func:`~mgkit.snps.mapper.map_taxon_id_to_rank`,
    with include_higher equals to False

    Arguments:
        snp_data (dict): dictionary sample->GeneSyn of SNPs data
        taxonomy: Uniprot Taxonomy
        min_num (int): minimum number of samples in which a valid pN/pS
            is found
        rank (str): taxon rank to map. Valid ranks are found in
            :data:`mgkit.taxon.TAXON_RANKS`
        index_type (str, None): type of index to return
        filters (iterable): list of filters to apply, otherwise uses the
            default filters

    Returns:
        DataFrame: :class:`pandas.DataFrame` of pN/pS values. The index type
        is 'taxon'
    """
    taxon_func = functools.partial(
        mgkit.snps.mapper.map_taxon_id_to_rank,
        taxonomy=taxonomy,
        rank=rank
    )

    if filters is None:
        filters = mgkit.snps.filter.get_default_filters(taxonomy)

    dataframe = mgkit.snps.funcs.combine_sample_snps(
        snp_data,
        min_num,
        filters,
        taxon_func=None if rank is None else taxon_func,
        gene_func=None,
        index_type=index_type
    )

    return dataframe


def get_gene_map_dataframe(snp_data, taxonomy, gene_map, min_num=3,
                           index_type='gene', filters=None):
    """
    .. versionadded:: 0.1.11

    .. versionchanged:: 0.2.2
        added *filters* argument

    Returns a :class:`~pandas.DataFrame` with the pN/pS of the given
    SNPs data, mapping all taxa to the gene map.

    Shortcut for using :func:`~mgkit.snps.funcs.combine_sample_snps`, using
    filters from :func:`~mgkit.snps.filter.get_default_filters` and as
    `gene_func` parameter :func:`~mgkit.snps.mapper.map_gene_id`.

    Arguments:
        snp_data (dict): dictionary sample->GeneSyn of SNPs data
        taxonomy: Uniprot Taxonomy
        min_num (int): minimum number of samples in which a valid pN/pS
            is found
        gene_map (dict): dictionary of mapping for the gene_ids in in SNPs
            data
        index_type (str, None): type of index to return
        filters (iterable): list of filters to apply, otherwise uses the
            default filters

    Returns:
        DataFrame: :class:`pandas.DataFrame` of pN/pS values. The index type
        is 'gene'
    """
    gene_func = functools.partial(
        mgkit.snps.mapper.map_gene_id,
        gene_map=gene_map
    )

    if filters is None:
        filters = mgkit.snps.filter.get_default_filters(taxonomy)

    dataframe = mgkit.snps.funcs.combine_sample_snps(
        snp_data,
        min_num,
        filters,
        taxon_func=None,
        gene_func=gene_func,
        index_type=index_type
    )

    return dataframe


def get_full_dataframe(snp_data, taxonomy, min_num=3, index_type=None,
                       filters=None):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.2.2
        added *filters* argument

    Returns a :class:`~pandas.DataFrame` with the pN/pS of the given
    SNPs data.

    Shortcut for using :func:`~mgkit.snps.funcs.combine_sample_snps`, using
    filters from :func:`~mgkit.snps.filter.get_default_filters`.

    Arguments:
        snp_data (dict): dictionary sample->GeneSyn of SNPs data
        taxonomy: Uniprot Taxonomy
        min_num (int): minimum number of samples in which a valid pN/pS
            is found
        index_type (str, None): type of index to return
        filters (iterable): list of filters to apply, otherwise uses the
            default filters

    Returns:
        DataFrame: :class:`pandas.DataFrame` of pN/pS values. The index type
        is None (gene-taxon)
    """

    if filters is None:
        filters = mgkit.snps.filter.get_default_filters(taxonomy)

    dataframe = mgkit.snps.funcs.combine_sample_snps(
        snp_data,
        min_num,
        filters,
        taxon_func=None,
        gene_func=None,
        index_type=index_type
    )

    return dataframe


def get_gene_taxon_dataframe(snp_data, taxonomy, gene_map, min_num=3,
                             rank='genus', index_type=None, filters=None,
                             use_uid=False):
    """
    .. versionadded:: 0.1.12

    .. versionchanged:: 0.2.2
        added *filters* argument

    .. versionchanged:: 0.5.1
        gene_map can be *None*, use_uid can be passed to the underline function

    .. todo::

        edit docstring

    Returns a :class:`~pandas.DataFrame` with the pN/pS of the given
    SNPs data, mapping all taxa to the gene map.

    Shortcut for using :func:`~mgkit.snps.funcs.combine_sample_snps`, using
    filters from :func:`~mgkit.snps.filter.get_default_filters` and as
    `gene_func` parameter :func:`~mgkit.snps.mapper.map_gene_id`.

    Arguments:
        snp_data (dict): dictionary sample->GeneSyn of SNPs data
        taxonomy: Uniprot Taxonomy
        min_num (int): minimum number of samples in which a valid pN/pS
            is found
        gene_map (dict): dictionary of mapping for the gene_ids in in SNPs
            data
        index_type (str, None): type of index to return
        filters (iterable): list of filters to apply, otherwise uses the
            default filters
        use_uid (bool): instead of using *gene_id*, uses *uid* as gene ID

    Returns:
        DataFrame: :class:`pandas.DataFrame` of pN/pS values. The index type
        is 'gene'
    """
    gene_func = functools.partial(
        mgkit.snps.mapper.map_gene_id,
        gene_map=gene_map
    )

    if rank is None:
        taxon_func = None
    else:
        taxon_func = functools.partial(
            mgkit.snps.mapper.map_taxon_id_to_rank,
            taxonomy=taxonomy,
            rank=rank
        )

    if filters is None:
        filters = mgkit.snps.filter.get_default_filters(taxonomy)

    dataframe = mgkit.snps.funcs.combine_sample_snps(
        snp_data,
        min_num,
        filters,
        taxon_func=taxon_func,
        gene_func=None if gene_map is None else gene_func,
        index_type=index_type,
        use_uid=use_uid
    )

    return dataframe



"""
SNPDat reader
"""
import logging
from builtins import object
import mgkit.utils.sequence
from mgkit.io import open_file

LOG = logging.getLogger(__name__)


class SNPDatRow(object):
    """
    Class containing information ouputted by SNPDat in its result file. One
    instance contains information about a row in the file.

    Attributes:
        chr_name (str): the queried SNPs chromosome ID
        chr_pos (int): queried SNPs genomic location
        in_feat (bool): Whether or not the queried SNP was within a feature
        region (str): Region containing the SNP; either exonic, intronic or
            intergenic
        feat_dist (int): Distance to nearest feature
        feature (str): Either the closest feature to the SNP or the feature
            containing the SNP
        num_features (int): number of different features that the SNP is
            annotated to
        feat_num (int): number of annotations of the current feature
        feat_start (int): Start of feature (bp)
        feat_end (int): End of feature (bp)
        gene_id (str): gene ID for the current feature
        gene_name (str): gene name for the current feature
        transcript_id (str): transcript ID for the current feature
        transcript_name (str): transcript name for the current feature
        exon (tuple): exon that contains the current feature and the total
            number of annotated exons for the gene containing the feature
        strand (str): strand sense of the feature
        ann_frame (str): annotated reading frame (when contained in the GTF)
        frame (str): reading frame estimated by SNPdat
        num_stops (int): estimated number of stop codons in the estimated
            reading frame
        codon (tuple): codon containing the SNP, position in the codon and
            reference base and mutation
        nuc_change (tuple): amino acid for the reference codon and new
            amino acid with the mutation in place
        nuc_ref (str, None): reference nucleotide
        aa_change (str): amino acid for the reference codon and new amino
            acid with the mutation in place
        synonymous (bool): Whether or not the mutation is synonymous
        protein_id (str): protein ID for the current feature
        messages (str): messages in the SNPDat line

    """
    __slots__ = (
        'chr_name',
        'chr_pos',
        'in_feat',
        'region',
        'feat_dist',
        'feature',
        'num_features',
        'feat_num',
        'feat_start',
        'feat_end',
        'gene_id',
        'gene_name',
        'transcript_id',
        'transcript_name',
        'exon',
        'strand',
        'ann_frame',
        'frame',
        'num_stops',
        'codon',
        'nuc_change',
        'nuc_ref',
        'aa_change',
        'synonymous',
        'protein_id',
        'messages'
    )

    def __init__(self, line=None, rev_comp=None):
        if rev_comp is None:
            rev_comp = mgkit.utils.sequence.REV_COMP
        if line:
            line = line.rstrip('\n').split('\t')
            self.chr_name = line[0].lower()  # The queried SNPs chromosome ID
            self.chr_pos = int(line[1])  # The queried SNPs genomic location
            # Whether or not the queried SNP was within a feature
            self.in_feat = True if line[2] == 'Y' else False
            # exonic - in a annotation - check overlapping genes
            # intergenic - between two features
            # NA - no annotation for that contig/snp
            # Region containing the SNP; either exonic, intronic or intergenic
            self.region = line[3].lower()
            # Distance to nearest feature
            if line[4] != 'NA':
                self.feat_dist = int(line[4].replace('*^', ''))
            else:
                self.feat_dist = 'NA'
            # Either the closest feature to the SNP or the feature containing
            # the SNP
            self.feature = line[5].lower()
            # The number of different features that the SNP is annotated to
            self.num_features = 0 if line[6] == 'NA' else int(line[6])
            # The number of annotations of the current feature
            if line[6] == 'NA':
                self.feat_num = 0
            else:
                tuple(int(x) for x in line[7][1:-1].split('/'))
            # Start of feature (bp)
            self.feat_start = 0 if line[6] == 'NA' else int(line[8])
            # End of feature (bp)
            self.feat_end = 0 if line[6] == 'NA' else int(line[9])
            self.gene_id = line[10]  # The gene ID for the current feature
            self.gene_name = line[11]  # The gene name for the current feature
            # The transcript ID for the current feature
            self.transcript_id = line[12]
            # The transcript name for the current feature
            self.transcript_name = line[13]
            # The exon that contains the current feature and the total number
            # of annotated exons for the gene containing the feature
            self.exon = tuple(x for x in line[14][1:-1].split('/'))
            # The strand sense of the feature
            self.strand = line[15].replace('*', '')
            # The annotated reading frame (when contained in the GTF)
            self.ann_frame = line[16]
            # The reading frame estimated by SNPdat
            self.frame = line[17].replace('*', '')
            # The estimated number of stop codons in the estimated reading
            # frame
            if line[6] == 'NA':
                self.num_stops = 0
            else:
                self.num_stops = int(line[18].replace('^', ''))
            # The codon containing the SNP, position in the codon and reference
            # base and mutation
            codon = line[19]
            if codon.startswith('['):
                # first base
                codon = (tuple(codon[1:-3].split('/')), codon[-2], codon[-1])
                if self.strand != '-':
                    self.nuc_change = codon[0][1]
                else:
                    self.nuc_change = rev_comp[codon[0][1]]
                if self.strand != '-':
                    self.nuc_ref = codon[0][0]
                else:
                    self.nuc_ref = rev_comp[codon[0][0]]
            elif codon.endswith(']'):
                # third base
                codon = (codon[0], codon[1], tuple(codon[3:-1].split('/')))
                if self.strand != '-':
                    self.nuc_change = codon[2][1]
                else:
                    self.nuc_change = rev_comp[codon[2][1]]
                if self.strand != '-':
                    self.nuc_ref = codon[2][0]
                else:
                    self.nuc_ref = rev_comp[codon[2][0]]
            elif codon == 'NA':
                codon = None
                self.nuc_change = None
                self.nuc_ref = None
            else:
                # second base
                codon = (codon[0], tuple(codon[2:-2].split('/')), codon[-1])
                if self.strand != '-':
                    self.nuc_change = codon[1][1]
                else:
                    self.nuc_change = rev_comp[codon[1][1]]
                if self.strand != '-':
                    self.nuc_ref = codon[1][0]
                else:
                    self.nuc_ref = rev_comp[codon[1][0]]
            self.codon = codon
            # The amino acid for the reference codon and new amino acid with
            # the mutation in place
            if line[20] == 'NA':
                self.aa_change = (None, None)
            else:
                self.aa_change = tuple(x for x in line[20][1:-1].split('/'))
            # Whether or not the mutation is synonymous
            self.synonymous = True if line[21] == 'Y' else False
            # The protein ID for the current feature
            self.protein_id = line[22]
            # not in a result file if no dbSNP ('-d' options) specified
            # self.rs_id = line[23] #The RS identifier for queries that map to
            # known SNPs
            self.messages = line[23]  # Error messages, warnings etc

    def __repr__(self):
        return str(self)

    def __str__(self):
        if self.codon is not None:
            codon = [
                x if isinstance(x, str) else '[{0}/{1}]'.format(*x)
                for x in self.codon
            ]
            codon = "{0}{1}{2}".format(*codon)
        else:
            codon = None
        return "{0}:{1} {2} - [{3}/{4}]".format(self.chr_name, self.chr_pos,
                                                codon, *self.aa_change)


def snpdat_reader(f_handle):
    """
    Simple SNPDat reader.

    f_handle: file handle or string for the SNPDat result file

    :return: generator of SNPDatRow instances
    """

    if isinstance(f_handle, str):
        f_handle = open_file(f_handle, 'r')
    LOG.info("Reading from file %s", f_handle.name)

    f_handle.readline()  # skips header line

    for line in f_handle:
        line = line.decode('ascii')
        try:
            yield SNPDatRow(line)
        except ValueError:
            LOG.critical(line)
            LOG.exception("Error reading line")
            raise ValueError



"""
Extract information from GFF files

sequence command
****************

Used to extract the nucleotidic sequences from GFF annotations. It requires the
*fasta* file containing the sequences referenced in the GFF `seq_id` attribute
(first column of the raw GFF).

The sequnces extract have as identifier the `uid` stored in the GFF file and by
default the sequnece is not reverse complemented if the annotation is on the
**-** strand, but this can be changed by using the `-r` option.

The sequences are wrapped at 60 characters, as per FASTA specs, but this
behavior can be disabled by specifing the `-w` option.

.. warning::

    The reference file is loaded in memory

dbm command
***********

Creates a dbm DB using the `semidbm` package. The database can then be loaded
using :class:`mgkit.db.dbm.GFFDB`

mongodb command
***************

Outputs annotations in a format supported by MongoDB. More information about it
can be found in :mod:`mgkit.db.mongo`

gtf command
***********

Outputs annotations in the GTF format

split command
*************

Splits a GFF file into smaller chunks, ensuring that all of a sequence
annotations are in the same file.

cov command
***********

Calculate annotation coverage for each contig in a GFF file. The command can be
run as strand specific (not by default) and requires the reference file to
which the annotation refer to. The output file is a tab separated one, with the
first column being the sequence name, the second is the strand (+, -, or NA if
not strand specific) and the third is the percentage of the sequence covered by
annotations.

.. warning::

    The GFF file is assumed to be sorted, by sequence or sequence-strand if
    wanted. The GFF file can be sorted using `sort -s -k 1,1 -k 7,7` for strand
    specific, or `sort -s -k 1,1` if not strand specific.

Changes
*******

.. versionchanged:: 0.3.4
    using *click* instead of *argparse*, renamed *split* command *--json* to
    *--json-out*

.. versionchanged:: 0.3.1
    added *cov* command

.. versionchanged:: 0.3.0
    added *--split* option to *sequence* command

.. versionchanged:: 0.2.6
    added *split* command, *--indent* option to *mongodb*

.. versionchanged:: 0.2.3
    added *--gene-id* option to *gtf* command

.. versionadded:: 0.2.2
    added *gtf* command

.. versionadded:: 0.2.1
    *dbm* and *mongodb* commands

.. versionadded:: 0.1.15

"""

from __future__ import division
import logging
import functools
import json
import click
from tqdm import tqdm
import mgkit
from . import utils
from mgkit.io import gff, fasta
from mgkit.db import dbm
from mgkit import taxon
from mgkit import simple_cache

LOG = logging.getLogger(__name__)


@click.group()
@click.version_option()
@utils.cite_option
def main():
    "Main function"
    pass


@main.command('sequence', help='''Extract the nucleotidic sequences of
annotations from [gff-file] to [fasta-file]''')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-r', '--reverse', is_flag=True,
              help='Reverse complement sequences on the - strand')
@click.option('-w', '--no-wrap', is_flag=True,
              help='Write the sequences on one line')
@click.option('-s', '--split', is_flag=True,
              help='''Split the sequence header of the reference at the first space, to emulate BLAST behaviour''')
@click.option('-f', '--reference', type=click.File('rb', lazy=False), default=None,
              help='Fasta file containing the reference sequences of the GFF file')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
@click.argument('fasta-file', type=click.File('wb', lazy=False), default='-')
def sequence_command(verbose, reverse, no_wrap, split, reference, progress,
                     gff_file, fasta_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    if reference is None:
        utils.exit_script('A fasta reference file is required', 1)

    wrap = 60

    if no_wrap:
        wrap = None

    seqs = dict(
        (
            seq_id.split(' ')[0] if split else seq_id,
            seq
        )
        for seq_id, seq in fasta.load_fasta(reference)
    )

    ann_iter = gff.parse_gff(gff_file, gff_type=gff.from_gff)

    seq_iter = gff.extract_nuc_seqs(ann_iter, seqs, reverse=reverse)

    if progress:
        seq_iter = tqdm(seq_iter)

    for name, seq in seq_iter:
        fasta.write_fasta_sequence(fasta_file, name, seq, wrap=wrap)


@main.command('dbm', help='''Creates a dbm database with annotations from file
[gff-file] into db [output-dir]''')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-d', '--output-dir', default='gff-dbm', show_default=True,
              help='Directory for the database')
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
def dbm_command(verbose, output_dir, gff_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    db = dbm.create_gff_dbm(gff.parse_gff(gff_file), output_dir)
    db.close()


@main.command('mongodb', help='''Extract annotations from a GFF [gff-file] file
and makes output for MongoDB [output-file]''')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-t', '--taxonomy', type=click.File('rb', lazy=False), default=None,
              help='Taxonomy used to populate the lineage')
@click.option('-c', '--no-cache', is_flag=True,
              help='No cache for the lineage function')
@click.option('-i', '--indent', type=click.INT, default=None,
              help='If used, the json will be written in a human readble form')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def mongodb_command(verbose, taxonomy, no_cache, indent, progress, gff_file,
                    output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(output_file, 'name', repr(output_file))
    )

    lineage_func = None

    if taxonomy is not None:
        taxonomy = taxon.Taxonomy(taxonomy)
        lineage_func = functools.partial(
            taxon.get_lineage,
            taxonomy
        )
        if no_cache:
            LOG.info('Using cached calls to lineage')
            lineage_func = simple_cache.memoize(lineage_func)

    iterator = gff.parse_gff(gff_file)

    if progress:
        iterator = tqdm(iterator)

    for annotation in iterator:
        output_file.write(
            annotation.to_mongodb(
                lineage_func=lineage_func,
                indent=indent
            ).encode('ascii')
        )
        output_file.write('\n'.encode('ascii'))


@main.command('gtf', help='''Extract annotations from a GFF file [gff-file] to
a GTF file [gtf-file]''')
@click.option('-v', '--verbose', is_flag=True)
@click.option('-g', '--gene-id', default='gene_id', show_default=True,
              help='GFF attribute to use for the GTF *gene_id* attribute')
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
@click.argument('gtf-file', type=click.File('wb', lazy=False), default='-')
def gtf_command(verbose, gene_id, gff_file, gtf_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        'Writing to file (%s)',
        getattr(gtf_file, 'name', repr(gtf_file))
    )

    for annotation in gff.parse_gff(gff_file):
        gtf_file.write(
            annotation.to_gtf(
                gene_id_attr=gene_id
            ).encode('ascii')
        )


@main.command('split', help="""Split annotations from a GFF file [gff-file] to
several files starting with [prefix]""")
@click.option('-v', '--verbose', is_flag=True,)
@click.option('-p', '--prefix', default='split', show_default=True,
              help='Prefix for the file name in output')
@click.option('-n', '--number', type=click.INT, default=10, show_default=True,
              help='Number of chunks into which split the GFF file')
@click.option('-z', '--gzip', is_flag=True, default=False,
              help='gzip output files')
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
def split_command(verbose, prefix, number, gzip, gff_file):
    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    LOG.info(
        "Splitting GFF into %d chunks with prefix %s",
        number,
        prefix
    )

    name_mask = "%s-{0:05}.gff" % prefix
    if gzip:
        name_mask += '.gz'
        LOG.info("Output files will be compressed (gzip)")

    gff.split_gff_file(
        gff_file,
        name_mask,
        num_files=number
    )


@main.command('cov', help="""Report on how much a sequence length is covered
by annotations in [gff-file]""")
@click.option('-v', '--verbose', is_flag=True)
@click.option('-f', '--reference', type=click.File('rb', lazy=False), required=True,
              help='Reference FASTA file for the GFF')
@click.option('-j', '--json-out', is_flag=True, default=False,
              help='The output will be a JSON dictionary')
@click.option('-s', '--strand-specific', is_flag=True, default=False,
              help='If the coverage must be calculated on each strand')
@click.option('-r', '--rename', default=False, is_flag=True,
              help='Emulate BLAST output (use only the header part before the first space)')
@click.option('--progress', default=False, is_flag=True,
              help="Shows Progress Bar")
@click.argument('gff-file', type=click.File('rb', lazy=False), default='-')
@click.argument('output-file', type=click.File('wb', lazy=False), default='-')
def coverage_command(verbose, reference, json_out, strand_specific, rename,
                     progress, gff_file, output_file):

    mgkit.logger.config_log(level=logging.DEBUG if verbose else logging.INFO)

    sequences = dict(
        fasta.load_fasta_rename(reference) if rename else fasta.load_fasta(reference)
    )
    iterator = gff.annotation_coverage_sorted(
        gff.parse_gff(gff_file),
        sequences,
        strand=strand_specific
    )
    if progress:
        iterator = tqdm(iterator)

    contig_coverage = {}

    for seq_id, strand, coverage in iterator:
        if json_out:
            contig_coverage[seq_id] = coverage
        else:
            output_file.write(
                "{}\t{}\t{}\n".format(
                    seq_id,
                    "NA" if strand is None else strand,
                    coverage
                ).encode('ascii')
            )

    if json_out:
        output_file.write(
            json.dumps(contig_coverage, indent=4).encode('ascii')
        )



"""
Utility functions for workflows
"""
from __future__ import print_function
import logging
import argparse
import sys
import click
import mgkit

LOG = logging.getLogger(__name__)


def cite_callback(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(mgkit.CITE)
    ctx.exit()


cite_option = click.option('--cite', is_flag=True, callback=cite_callback,
                           expose_value=False, is_eager=True)


class CiteAction(argparse.Action):
    """
    Argparse action to print the citation, using the :func:`mgkit.cite`
    function.
    """
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help='Show citation for the framework'):
        super(CiteAction, self).__init__(
            option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)

    def __call__(self, parser, namespace, values, option_string=None):
        mgkit.cite(file_handle=sys.stderr)
        setattr(namespace, self.dest, values)
        parser.exit(0)


class PrintManAction(argparse.Action):
    """
    .. versionadded:: 0.2.6

    Argparse action to print the manual
    """
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 help='Show the script manual',
                 manual=''):
        super(PrintManAction, self).__init__(
            option_strings,
            dest=dest,
            default=default,
            nargs=0,
            help=help)
        self.manual = manual

    def __call__(self, parser, namespace, values, option_string=None):
        print(self.manual, file=sys.stdout)
        setattr(namespace, self.dest, values)
        parser.exit(0)


def add_basic_options(parser, manual=''):
    """
    .. versionchanged:: 0.2.6
        added *quiet* option

    Adds verbose and version options to the option parser
    """
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '-v',
        '--verbose',
        action='store_const',
        const=logging.DEBUG,
        default=logging.INFO,
        help='more verbose - includes debug messages',
        dest='verbose'
    )
    group.add_argument(
        '--quiet',
        action='store_const',
        const=logging.ERROR,
        help='less verbose - only error and critical messages',
        dest='verbose'
    )
    parser.add_argument(
        '--cite',
        action=CiteAction
    )
    parser.add_argument(
        '--manual',
        action=PrintManAction,
        manual=manual
    )
    parser.add_argument(
        '--version',
        action='version',
        version='%(prog)s {0}'.format(mgkit.__VERSION__)
    )


def exit_script(message, ret_value):
    """
    Used to exit the script with a return value
    """
    LOG.critical(message)
    sys.exit(ret_value)



"""
Metagenomics Framework
"""
from __future__ import print_function
import sys
import logging

from . import logger

__VERSION__ = "0.5.3"

__version__ = __VERSION__

LOG = logging.getLogger(__name__)

DEBUG = False
"Debug switch for a few functions"

LOGO = """
 _|      _|    _|_|_|  _|    _|  _|    _|
 _|_|  _|_|  _|        _|  _|        _|_|_|_|
 _|  _|  _|  _|  _|_|  _|_|      _|    _|
 _|      _|  _|    _|  _|  _|    _|    _|
 _|      _|    _|_|_|  _|    _|  _|      _|_|

"""

CITE = """
Rubino, F. and Creevey, C.J. (2014).
MGkit: Metagenomic Framework For The Study Of Microbial Communities.

Available at: https://bit.ly/2KUunnN

[doi:10.6084/m9.figshare.1269288]
"""

PKG_NAME = 'MGKit'


def cite(file_handle=sys.stderr):
    """
    Print citation to the specified stream
    """
    print(
        LOGO,
        'MGKit Version: {0}'.format(__VERSION__),
        CITE,
        sep='\n',
        file=file_handle,
        end=''
    )


def check_version(version):
    if __version__ != version:
        LOG.warning(
            "This was tested with %s version %s (%s was found)",
            PKG_NAME,
            version,
            __version__
        )


class DependencyError(Exception):
    "Raised if an optional requirement is needed"
    def __init__(self, package):
        super(DependencyError, self).__init__(
            "The '{}' package is required".format(package)
        )



import pytest

from mgkit.utils.common import average_length, between, union_range, \
    range_intersect, union_ranges, complement_ranges, ranges_length, \
    apply_func_window


@pytest.mark.parametrize(
    "coords,result",
    [
        ((1, 200, 3, 132), 165.0),
        ((4, 2245, 3, 13223), 7731.5),
    ]
)
def test_avg_len(coords, result):
    assert average_length(*coords) == result


@pytest.mark.parametrize(
    "coords,result",
    (
        ((1, 0, 10), True),
        ((1, 2, 10), False),
        ((1, 1, 10), True),
        ((11, 1, 10), False),
        ((10, 1, 10), True),
    )
)
def test_between(coords, result):
    assert between(*coords) == result


@pytest.mark.parametrize(
    "coords,result",
    [
        ((1, 10, 2, 13), (1, 13)),
        ((1, 10, 10, 13), (1, 13)),
        ((10, 13, 1, 10), (1, 13)),
        ((1, 10, 11, 13), (1, 13)),
        ((10.0, 13.0, 1, 10), (1, 13)),
    ]
)
def test_union_range(coords, result):
    assert union_range(*coords) == result


@pytest.mark.parametrize(
    "coords,result",
    [
        ((10, 20, 19, 30), (19, 20)),
        ((10, 20, 15, 30), (15, 20)),
        ((10, 20, 10, 20), (10, 20)),
        ((10, 20, 12, 18), (12, 18)),
        # The next one is to test the result for no intersection
        ((10, 20, 30, 40), None)
    ]
)
def test_range_intersect(coords, result):
    assert range_intersect(*coords) == result


@pytest.mark.parametrize(
    "intervals,result",
    [
        ([(1, 2), (3, 7), (6, 12), (9, 17), (18, 20)], [(1, 20)]),
        ([(3, 7), (6, 12), (9, 17), (18, 20), (1, 2)], [(1, 20)]),
        ([(1, 2), (3, 7), (6, 12), (9, 14), (18, 20)], [(1, 14), (18, 20)])
    ]
)
def test_union_ranges(intervals, result):
    assert union_ranges(intervals) == result


@pytest.mark.parametrize(
    "intervals,end,result",
    [
        ([(1, 10), (11, 20), (25, 30)], 100, [(21, 24), (31, 100)]),
        ([(1, 10), (11, 20), (25, 30)], None, [(21, 24)]),
        ([(0, 2), (3, 17), (18, 20)], None, []),
        ([(0, 2), (3, 17), (18, 20)], 100, [(21, 100)]),
        ([(3, 17), (18, 20)], 100, [(1, 2), (21, 100)])
    ]
)
def test_complement_ranges(intervals, end, result):
    assert complement_ranges(intervals, end) == result


@pytest.mark.parametrize(
    "coords,result",
    [
        ([(1, 10), (12, 25), (25, 30)], 30),
        # They overlap by one,, so the sum will be 26, instead of 25
        # Tested to make sure we can rely on this behaviour
        # as per warning in the documentation
        ([(1, 10), (10, 25)], 26),
    ]
)
def test_ranges_length(coords, result):
    assert ranges_length(coords) == result


@pytest.mark.parametrize(
    "func,data,window,step,result",
    [
        [sum, range(10), 10, 0, [45]],
        [sum, range(10), 10, 9, [45, 9]],
        [sum, range(10), 5, 5, [10, 35]],
    ]
)
def test_apply_func_window(func, data, window, step, result):
    reduction = list(
        apply_func_window(func, data, window, step)
    )

    assert reduction == result



import pytest
import pandas
import numpy
from mgkit.counts.scaling import scale_factor_deseq, scale_deseq


def test_scale_scale_factor_deseq():
    array = numpy.random.negative_binomial(100, .01, size=1000)
    dataframe = pandas.DataFrame(
        {1: array, 2: array * 1.5, 3: array * 2}
    )
    sf = scale_factor_deseq(dataframe)
    assert sf[1] < sf[2] < sf[3]


def test_scale_scale_deseq():
    array = numpy.random.negative_binomial(100, .01, size=1000)
    dataframe = pandas.DataFrame(
        {1: array, 2: array * 1.5, 3: array * 2}
    )
    sdf = scale_deseq(dataframe)
    sf = scale_factor_deseq(dataframe)
    assert (sdf * sf).round(0).iloc[0, :].sum() == dataframe.round(0).iloc[0, :].sum()



import pytest
from conftest import skip_no_connection, taxonomy_files
from mgkit.taxon import TaxonTuple, Taxonomy, TAXON_RANKS, NoLcaFound, \
    last_common_ancestor, last_common_ancestor_multiple, VIRUS, \
    CELLULAR_ORGANISMS


@pytest.fixture
def taxonomy():
    t = Taxonomy()
    for x in range(10):
         lineage = {'{}_'.format(rank): '{}{}'.format(rank, x) for rank in TAXON_RANKS}
         t.add_lineage(**lineage)

    return t


def test_taxonomy_add_lineage1():
    t = Taxonomy()
    lineage = {'{}_'.format(rank): '{}{}'.format(rank, 1) for rank in TAXON_RANKS}
    t.add_lineage(**lineage)


def test_taxonomy_add_lineage2():
    t = Taxonomy()
    lineage = {'{}_'.format(rank): '{}{}'.format(rank, 1) for rank in TAXON_RANKS}
    lineage['subspecies'] = 'subspecies1'
    t.add_lineage(**lineage)


def test_taxonomy_add_lineage3():
    t = Taxonomy()
    lineage = {'{}_'.format(rank): '{}{}'.format(rank, 1) for rank in TAXON_RANKS}
    lineage['subspecies'] = 'subspecies1'
    lineage['cultivar'] = 'cultivar1'
    with pytest.raises(ValueError):
        t.add_lineage(**lineage)


def test_taxonomy_get_lineage_string1(taxonomy):
    taxonomy.get_lineage_string(27) == 'superkingdom2;kingdom2;phylum2;class2;subclass2;order2;family2;genus2;species2'


def test_taxonomy_get_lineage_string2(taxonomy):
    taxonomy.get_lineage_string(27, rank='order') == 'superkingdom2;kingdom2;phylum2;class2;subclass2;order2'


# These  zill generate the name map
def test_taxonomy_find_by_name_base(taxonomy):
    name = 'species3'
    assert taxonomy.find_by_name(name) == [36]


def test_taxonomy_find_by_name_rank1(taxonomy):
    name = 'species3'
    assert taxonomy.find_by_name(name, rank='species') == 36


def test_taxonomy_find_by_name_rank2(taxonomy):
    name = 'species3'
    assert taxonomy.find_by_name(name, rank='genus') == None


def test_taxonomy_find_by_name_rank3():
    taxonomy = Taxonomy()
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p1', order='test_o')
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p2', order='test_o')
    with pytest.raises(KeyError):
        taxon_id = taxonomy.find_by_name('test_o', rank='order')


def test_taxonomy_find_by_name_rank4():
    taxonomy = Taxonomy()
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p1', order='test_o')
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p2', order='test_o')
    taxon_id = taxonomy.find_by_name('test_o', rank='order', strict=False)
    assert len(taxon_id) == 2


def test_taxonomy_add_taxon1():
    taxonomy = Taxonomy()
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p1', order='test_o')
    taxon_id = taxonomy.add_lineage(superkingdom='test_k', phylum='test_p2', order='test_o')
    parent_id = taxonomy.find_by_name('test_p2', rank='phylum')
    assert taxon_id == taxonomy.add_taxon('test_o', rank='order', parent_id=parent_id)


def test_taxonomy_add_taxon2():
    taxonomy = Taxonomy()
    taxonomy.add_lineage(superkingdom='test_k', phylum='test_p1', order='test_o')
    taxon_id = taxonomy.add_lineage(superkingdom='test_k', phylum='test_p2', order='test_o')
    with pytest.raises(KeyError):
        taxonomy.add_taxon('test_o', rank='order', parent_id=1)


def test_taxonomy_isancestor1(taxonomy):
    assert taxonomy.is_ancestor(9, 1)


def test_taxonomy_isancestor2(taxonomy):
    assert not taxonomy.is_ancestor(10, 1)


def test_taxonomy_isancestor3(taxonomy):
    assert not taxonomy.is_ancestor(10**4, 1)


def test_taxonomy_drop_taxon(taxonomy):
    taxonomy.drop_taxon(6)
    assert not (8 in taxonomy)


def test_taxonomy_drop_taxon_fail(taxonomy):
    taxonomy.drop_taxon(6)
    with pytest.raises(KeyError):
        taxonomy.get_lineage_string(9)


def test_taxonomy_get_ranked_taxon1(taxonomy):
    assert taxonomy.get_ranked_taxon(9, rank='family').taxon_id == 7


def test_taxonomy_get_ranked_taxon2(taxonomy):
    with pytest.raises(ValueError):
        taxonomy.get_ranked_taxon(9., rank='family')


def test_last_common_ancestor(taxonomy):
    lca = 5
    other = taxonomy.add_taxon('test-lca', parent_id=lca)
    assert last_common_ancestor(taxonomy, 9, other) == lca


def test_last_common_ancestor_fail(taxonomy):
    with pytest.raises(NoLcaFound):
        last_common_ancestor(taxonomy, 9, 10)


def test_last_common_ancestor_multiple(taxonomy):
    lca = 5
    other1 = taxonomy.add_taxon('test-lca', parent_id=lca)
    other2 = taxonomy.add_taxon('test-lca', parent_id=7)
    assert last_common_ancestor_multiple(taxonomy, [9, other1, other2]) == lca


def test_last_common_ancestor_multiple_fail1(taxonomy):
    with pytest.raises(NoLcaFound):
        last_common_ancestor_multiple(taxonomy, [])


def test_last_common_ancestor_multiple_fail2(taxonomy):
    with pytest.raises(NoLcaFound):
        last_common_ancestor_multiple(taxonomy, [5, 8, 10])


def test_taxonomy_serialise_pickle(taxonomy, tmpdir):
    file_name = tmpdir.join('tx.pickle').strpath

    taxonomy.save_data(file_name)

    tx2 = Taxonomy(file_name)

    assert taxonomy._taxa == tx2._taxa


def test_taxonomy_serialise_msgpack(taxonomy, tmpdir):
    file_name = tmpdir.join('tx.msgpack').strpath

    taxonomy.save_data(file_name)

    tx2 = Taxonomy(file_name)

    assert taxonomy._taxa == tx2._taxa


@skip_no_connection
def test_read_ncbi_taxonomy(taxonomy_files):

    taxonomy = Taxonomy()
    taxonomy.read_from_ncbi_dump(*taxonomy_files)

    assert taxonomy.find_by_name('prevotella', rank='genus') == 838


@skip_no_connection
def test_root_taxa1(ncbi_taxonomy):
    genus1 = ncbi_taxonomy.find_by_name('prevotella', rank='genus')
    genus2 = ncbi_taxonomy.find_by_name('methanobrevibacter', rank='genus')
    assert last_common_ancestor(ncbi_taxonomy, genus1, genus2) == CELLULAR_ORGANISMS


@skip_no_connection
def test_root_taxa2(ncbi_taxonomy):
    genus1 = ncbi_taxonomy.find_by_name('prevotella', rank='genus')
    with pytest.raises(NoLcaFound):
        last_common_ancestor(ncbi_taxonomy, genus1, VIRUS)


def test_taxonomy_parse_gtdb_lineage():
    lineage = 'd__kingdom1;p__phylum1;o__order1;c__class1;f__family1;g__genus1;s__species1'

    result_dict = Taxonomy.parse_gtdb_lineage(lineage)

    assert result_dict == {'superkingdom': 'kingdom1', 'phylum': 'phylum1', 'order': 'order1', 'class': 'class1', 'family': 'family1', 'genus': 'genus1', 'species': 'species1'}


def test_taxonomy_get_ranked_id1(taxonomy):
    assert taxonomy.get_ranked_id(9, rank='family') == 7


def test_taxonomy_get_ranked_id2(taxonomy):
    assert taxonomy.get_ranked_id(2, rank='family', include_higher=True) == 2


def test_taxonomy_get_ranked_id3(taxonomy):
    assert taxonomy.get_ranked_id(2, rank='family', include_higher=False) is None


def test_taxonomy_get_ranked_id4(taxonomy):
    assert taxonomy.get_ranked_id(9, rank='family', it=True) == [7]


def test_taxonomy_delitem(taxonomy):
    del taxonomy[9]
    assert 9 not in taxonomy


def test_taxonomy_len(taxonomy):
    assert len(taxonomy) == 90


def test_taxonomy_contains(taxonomy):
    assert 10000 not in taxonomy


def test_taxonomy_repr(taxonomy):
    assert repr(taxonomy) == "<class 'mgkit.taxon.Taxonomy'> - 90 taxa"
